# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import traceback
import math
import struct
import time
import os.path
import re

import  numpy

import Kittens.utils

import Tigger.Models.Formats
from Tigger.Models import ModelClasses
from Tigger.Models import SkyModel
from Tigger import Coordinates
from Tigger.Models.Formats import dprint,dprintf


"""
The BBS sky model catalog file (*.cat, or *.catalog) is a human-readable text
file that contains a list of sources. The file should be in the `makesourcedb'
format. For details, please refer to
http://www.lofar.org/operations/doku.php?id=engineering:software:tools:makesourcedb#format_string
or
http://www.lofar.org/operations/doku.php?id=engineering:software:tools:bbs#creating_a_catalog_file
"""

class CatalogLine (object):
  """A CatalogLine turns one catalog file line into an object whose attributes correspond to the fields.
  """;
  def __init__ (self,parser,fields=None):
    """Creates a catalog line. If fields!=None, then this contains a list of fields already filled in""";
    self._parser = parser;
    self._fields = fields;
    if fields:
      # parse fields
      for field,number in parser.field_number.iteritems():
        fval = fields[number].strip() if number < len(fields) else '';
        if not fval:
          fval = parser.field_default.get(field,'');
        setattr(self,field,fval);
      # make directions
      self.ra_rad = parser.getAngle(self,'Ra','rah','rad','ram','ras');
      self.dec_rad = parser.getAngle(self,'Dec','dech','decd','decm','decs');
    else:
      # else make empty line
      for field in parser.field_number.iterkeys():
        setattr(self,field,'');
        
  def setPosition (self,ra,dec):
    """Sets the position ra/dec in radians: fills in fields according to the parser format""";
    self.ra_rad,self.dec_rad = ra,dec;
    self._parser.putAngle(self,ra,'Ra','rah','rad','ram','ras');
    self._parser.putAngle(self,dec,'Dec','dech','decd','decm','decs');
        
  def makeStr (self):
    """Converts into a string using the designated parser""";
    # build up dict of valid fields
    fields = {};
    for field,num in self._parser.field_number.iteritems():
      value = getattr(self,field,None);
      if value:
        fields[num] = value;
    # output
    output = "";
    nfields = max(fields.iterkeys())+1;
    for i in range(nfields):
      sep = self._parser.separators[i] if i<nfields-1 else '';
      output += "%s%s"%(fields.get(i,''),sep);
    return output;
  
class CatalogParser (object):
  def __init__ (self,format):
    # figure out fields and their separators
    fields = [];
    self.separators = [];
    while True:
      match = re.match("(\w[\w:]*(=(fixed)?'[^']*')?)(([^\w]+)(\w.*))?$",format);
      if not match:
        break;
      fields.append(match.group(1));
      # if no group 4, then we've reached the last field
      if not match.group(4):
        break;
      self.separators.append(match.group(5));
      format = match.group(6);
    # now parse the format specification
    # this is a dict of field name -> field index
    self.field_number = {};
    # this is a dict of field name -> default value
    self.field_default = dict(Category='2',I='1');
    # fill up the dicts
    for num_field,field in enumerate(fields):
      # is a default value given?
      match = re.match("(.+)='(.*)'$",field);
      if match:
        field = match.group(1);
        self.field_default[field] = match.group(2);
      self.field_number[field] = num_field;
    dprint(2,"fields are",self.field_number);
    dprint(2,"default values are",self.field_default);
    dprint(2,"separators are",self.separators);
    
  def defines (self,field):
    return field in self.field_number;
    
  def parse (self,line,linenum=0):
    """Parses one line. Returns None for empty or commented lines, else returns a CatalogLine object""";
    # strip whitespace
    line = line.strip();
    dprintf(3,"read line %d: %s\n",linenum,line);
    # skip empty or commented lines
    if not line or line[0] == '#':
      return None;
    # split using separators, quit when no more separators
    fields = [];
    for sep in self.separators:
      ff = line.split(sep,1);
      if len(ff) < 2:
        break;
      fields.append(ff[0]);
      line = ff[1];
    fields.append(line);
    dprint(4,"line %d: "%linenum,fields);
    return CatalogLine(self,fields);
    
  def newline (self):
    return CatalogLine(self);

  def getAngle (self,catline,field,fh,fd,fm,fs):
    """Helper function: given a CatalogLine, and a set of field indentifiers, turns this
    into an angle (in radians).""";
    scale = 1;
    if self.defines(field):
      fstr = getattr(catline,field,None);
      match = re.match('([+-]?\s*\d+)[h:](\d+)[m:]([\d.]*)s?$',fstr);
      if match:
        scale = 15;
      else:
        match = re.match('([+-]?\s*\d+).(\d+).(.*)$',fstr);
        if not match:
          raise ValueError,"invalid direction '%s'"%fstr;
      d,m,s = match.groups();
    else:
      if self.defines(fh):
        scale = 15;
        d = getattr(catline,fh);
      else:
        d = getattr(catline,fd,'0');
      m = getattr(catline,fm,'0');
      s = getattr(catline,fs,'0');
    # now, d,m,s are strings
    if d.startswith('-'):
      scale = -scale;
      d = d[1:];
    # convert to degrees
    return scale*(float(d) + float(m)/60 + float(s)/3600)*math.pi/180;

  def putAngle (self,catline,angle,field,fh,fd,fm,fs,prec=1e-6):
    """Helper function: inverse of getAngle.""";
    # decompose angle into sign,d,m,s
    if angle < 0:
      sign = "-";
      angle = -angle;
    else:
      sign = "+" if field == "Dec" else "";
    angle *= 12/math.pi if not self.defines(field) and self.defines(fh) else 180/math.pi;
    mins,secs = divmod(round(angle*3600/prec)*prec,60);
    mins = int(mins);
    degs,mins = divmod(mins,60);
    #generate output
    if self.defines(field):
      setattr(catline,field,"%s%d.%d.%.4f"%(sign,degs,mins,secs));
    else:
      setattr(catline,fh if self.defines(fh) else fd,"%s%d"%(sign,degs));
      setattr(catline,fm,"%d"%mins);
      setattr(catline,fs,"%.4f"%secs);


def load (filename,freq0=None,center_on_brightest=False,**kw):
  """Imports an BBS catalog file
  The 'format' argument can be either a dict (such as the DefaultDMSFormat dict above), or a string such as DefaultDMSFormatString.
  (Other possible field names are "ra_d", "ra_rad", "dec_rad", "dec_sign".)
  If None is specified, DefaultDMSFormat is used.
  The 'freq0' argument supplies a default reference frequency (if one is not contained in the file.)
  If 'center_on_brightest' is True, the mpodel field center will be set to the brightest source,
  else to the center of the first patch.
  """
  srclist = [];
  dprint(1,"importing BBS source table",filename);
  # read file
  ff = file(filename);
  # first line must be a format string: extract it
  line0 = ff.readline().strip();
  match = re.match("#\s*\((.+)\)\s*=\s*format",line0);
  if not match:
    raise ValueError,"line 1 is not a valid format specification";
  format_str = match.group(1);
  # create format parser from this string
  parser = CatalogParser(format_str);
    
  # check for mandatory fields
  for field in "Name","Type":
    if not parser.defines(field):
      raise ValueError,"Table lacks mandatory field '%s'"%field;

  maxbright = 0;
  patches = [];
  ref_freq = freq0;

  # now process file line-by-line
  linenum = 1;
  for line in ff:
    linenum += 1;
    try:
      # parse one line
      dprint(4,"read line:",line);
      catline = parser.parse(line,linenum);
      if not catline:
        continue;
      dprint(5,"line %d: "%linenum,catline.__dict__);
      # is it a patch record?
      patchname = getattr(catline,'Patch','');
      if not catline.Name:
        dprintf(2,"%s:%d: patch %s\n",filename,linenum,patchname);
        patches.append((patchname,catline.ra_rad,catline.dec_rad));
        continue;
      # form up name
      name = "%s:%s"%(patchname,catline.Name) if patchname else catline.Name;
      # check source type
      stype = catline.Type.upper();
      if stype not in ("POINT","GAUSSIAN"):
        raise ValueError,"unsupported source type %s"%stype;
      # see if we have freq0
      if freq0:
        f0 = freq0;
      elif hasattr(catline,'ReferenceFrequency'):
        f0 = float(catline.ReferenceFrequency or '0');
      else:
        f0 = None;
      # set model refrence frequency
      if f0 is not None and ref_freq is None:
        ref_freq = f0;
      # see if we have Q/U/V
      i,q,u,v = [ float(getattr(catline,stokes,'0') or '0') for stokes in "IQUV" ];
      # see if we have RM as well. Create flux object (unpolarized, polarized, polarized w/RM)
      if f0 is not None and hasattr(catline,'RotationMeasure'):
        flux = ModelClasses.PolarizationWithRM(i,q,u,v,float(catline.RotationMeasure or '0'),f0);
      else:
        flux = ModelClasses.Polarization(i,q,u,v);
      # see if we have a spectral index
      if f0 is not None and hasattr(catline,'SpectralIndex:0'):
        spectrum = ModelClasses.SpectralIndex(float(getattr(catline,'SpectralIndex:0') or '0'),f0);
      else:
        spectrum = None;
      # see if we have extent parameters
      if stype == "GAUSSIAN":
        ex = float(getattr(catline,"MajorAxis","0") or "0");
        ey = float(getattr(catline,"MinorAxis","0") or "0");
        pa = float(getattr(catline,"Orientation","0") or "0");
        shape = ModelClasses.Gaussian(ex,ey,pa);
      else:
        shape = None;
      # create tags
      tags = {};
      for field in "Patch","Category":
        if hasattr(catline,field):
          tags['BBS_%s'%field] = getattr(catline,field);
      # OK, now form up the source object
      # position
      pos = ModelClasses.Position(catline.ra_rad,catline.dec_rad);
      # now create a source object
      src = SkyModel.Source(name,pos,flux,shape=shape,spectrum=spectrum,**tags);
      srclist.append(src);
      # check if it's the brightest
      brightness = src.brightness();
      if brightness > maxbright:
        maxbright = brightness;
        brightest_name = src.name;
        radec0 = catline.ra_rad,catline.dec_rad;
    except:
      dprintf(0,"%s:%d: %s, skipping\n",filename,linenum,str(sys.exc_info()[1]));
  dprintf(2,"imported %d sources from file %s\n",len(srclist),filename);
  # create model
  model = ModelClasses.SkyModel(*srclist);
  if ref_freq is not None:
    model.setRefFreq(ref_freq);
  # setup model center
  if center_on_brightest and radec0:
    dprintf(2,"setting model centre to brightest source %s (%g Jy) at %f,%f\n",brightest_name,maxbright,*radec0);
    model.setFieldCenter(*radec0);
  elif patches:
    name,ra,dec = patches[0];
    dprintf(2,"setting model centre to first patch %s at %f,%f\n",name,ra,dec);
    model.setFieldCenter(ra,dec);
  # map patches to model tags
  model.setAttribute("BBS_Patches",patches);
  model.setAttribute("BBS_Format",format_str);
  # setup radial distances
  projection = Coordinates.Projection.SinWCS(*model.fieldCenter());
  for src in model.sources:
    l,m = projection.lm(src.pos.ra,src.pos.dec);
    src.setAttribute('r',math.sqrt(l*l+m*m));
  return model;

def save (model,filename,sources=None,format=None,**kw):
  """Exports model to a BBS catalog file""";
  if sources is None:
    sources = model.sources;
  dprintf(2,"writing %d model sources to BBS file %s\n",len(sources),filename);
  # create catalog parser based on either specified format, or the model format, or the default format
  format = format or getattr(model,'BBS_Format',
        "Name, Type, Patch, Ra, Dec, I, Q, U, V, ReferenceFrequency, SpectralIndexDegree='0', SpectralIndex:0='0.0', MajorAxis, MinorAxis, Orientation");
  dprint(2,"format string is",format);
  parser = CatalogParser(format);
  # check for mandatory fields
  for field in "Name","Type":
    if not parser.defines(field):
      raise ValueError,"Output format lacks mandatory field '%s'"%field;
  # open file
  ff = open(filename,mode="wt");
  ff.write("# (%s) = format\n# The above line defines the field order and is required.\n\n"%format);
  # write patches
  for name,ra,dec in getattr(model,"BBS_Patches",[]):
    catline = parser.newline();
    catline.Patch = name;
    catline.setPosition(ra,dec);
    ff.write(catline.makeStr()+"\n");
  ff.write("\n");
  # write sources
  nsrc = 0;
  for src in sources:
    catline = parser.newline();
    # type
    if src.shape is None:
      catline.Type = "POINT";
    elif isinstance(src.shape,ModelClasses.Gaussian):
      catline.Type = "GAUSSIAN";
    else:
      dprint(3,"skipping source '%s': non-supported type '%s'"%(src.name,src.shape.typecode));
      continue;
    # name and patch
    name = src.name;
    patch = getattr(src,'BBS_Patch','');
    if patch and name.startswith(patch+':'):
      name = name[(len(patch)+1):]
    catline.Name = name;
    setattr(catline,'Patch',patch);
    # position
    catline.setPosition(src.pos.ra,src.pos.dec);
    # fluxes
    for stokes in "IQUV":
      setattr(catline,stokes,str(getattr(src.flux,stokes,0.)));
    # reference freq
    freq0 = (src.spectrum and getattr(src.spectrum,'freq0',None)) or getattr(src.flux,'freq0',None);
    if freq0 is not None:
      setattr(catline,'ReferenceFrequency',str(freq0));
    # RM, spi
    if isinstance(src.spectrum,ModelClasses.SpectralIndex):
      setattr(catline,'SpectralIndexDegree','0');
      setattr(catline,'SpectralIndex:0',str(src.spectrum.spi));
    if isinstance(src.flux,ModelClasses.PolarizationWithRM):
      setattr(catline,'RotationMeasure',str(src.flux.rm));
    # shape
    if isinstance(src.shape,ModelClasses.Gaussian):
      setattr(catline,'MajorAxis',src.shape.ex);
      setattr(catline,'MinorAxis',src.shape.ey);
      setattr(catline,'Orientation',src.shape.pa);
    # write line
    ff.write(catline.makeStr()+"\n");
    nsrc += 1;
    
  ff.close();
  dprintf(1,"wrote %d sources to file %s\n",nsrc,filename);


Tigger.Models.Formats.registerFormat("BBS",load,"BBS source catalog",(".cat",".catalog"),export_func=save);



# -*- coding: utf-8 -*-

#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from Timba.TDL import *
from Timba.utils import curry
import traceback
import Meow
import Meow.OptionTools
import Meow.Context
import Meow.ParmGroup
import math
from math import *
import os.path

from Meow.MeqMaker import SourceSubsetSelector

# find out where Tigger lives -- either it's in the path, or we add it
try:
  import Tigger
except:
  sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))));
  import Tigger

from Tigger.Models import ModelClasses
from Tigger.Models.Formats import ModelHTML

# this dict determines how source attributes are grouped into "parameter subgroups"
_Subgroups = dict(I="I",Q="Q",U="U",V="V",
    ra="pos",dec="pos",RM="RM",spi="spi",
    sx="shape",sy="shape",phi="shape");
_SubgroupOrder = "I","Q","U","V","pos","spi","RM","shape";

class TiggerSkyModel (object):
  """Interface to a Tigger-format sky model."""
  def __init__ (self,filename=None,include_options=False,tdloption_namespace='tiggerlsm'):
    """Initializes a TiggerSkyModel object.
    A filename and a format may be specified, although the actual file will
    only be loaded on demand.
    If include_options=True, immediately instantiates the options. If False, it is up to
    the caller to include the options in his menus.
    """;
    self.tdloption_namespace = tdloption_namespace;
    self._compile_opts = [];
    self._runtime_opts = [];
    self.filename = filename;
    self.lsm = None;
    # immediately include options, if needed
    if include_options:
      TDLCompileOptions(*self.compile_options());
      TDLRuntimeOptions(*self.runtime_options());

  def compile_options (self):
    """Returns list of compile-time options""";
    if not self._compile_opts:
      self._compile_opts = [
        TDLOption("filename","Tigger LSM file",
                   TDLFileSelect("Tigger models (*."+ModelHTML.DefaultExtension+");;All files (*)",default=self.filename,exist=True),
                   namespace=self),
        TDLOption('lsm_subset',"Source subset",["all"],more=str,namespace=self,
                  doc=SourceSubsetSelector.docstring),
        TDLOption('null_subset',"Use nulls for subset",[None],more=str,namespace=self,doc=
                  """<P>If you wish, any subset of sources may be "nulled" by inserting a null
                  brightness for them. This is used in some advanced calibration scenarios; if
                  you're not sure about this option, just leave it set to "None".</P>
                  </P>"""+SourceSubsetSelector.docstring),
        TDLMenu("Make solvable source parameters",
          TDLOption('lsm_solvable_tag',"Solvable source tag",[None,"solvable"],more=str,namespace=self,
                    doc="""If you specify a tagname, only sources bearing that tag will be made solvable. Use 'None' to make all sources solvable."""),
          TDLOption('lsm_solve_group_tag',"Group independent solutions by tag",[None,"cluster"],more=str,namespace=self,
                    doc="""If you specify a tagname, sources will be grouped by the value of the tag,
                    and each group will be treated as an independent solution."""),
          TDLOption("solve_I","I",False,namespace=self),
          TDLOption("solve_Q","Q",False,namespace=self),
          TDLOption("solve_U","U",False,namespace=self),
          TDLOption("solve_V","V",False,namespace=self),
          TDLOption("solve_spi","spectral index",False,namespace=self),
          TDLOption("solve_pos","position",False,namespace=self),
          TDLOption("solve_RM","rotation measure",False,namespace=self),
          TDLOption("solve_shape","shape (for extended sources)",False,namespace=self),
          toggle='solvable_sources',namespace=self,
        )
    ];
    return self._compile_opts;

  def runtime_options (self):
    """Makes and returns list of compile-time options""";
    # no runtime options, for now
    return self._runtime_opts;
    
  # helper function for use with SourceSubsetSelector below
  @staticmethod
  def _getTagValue (src,tag):
    """Helper function: looks for the given tag in the source, or in its sub-objects""";
    for obj in src,src.pos,src.flux,getattr(src,'shape',None),getattr(src,'spectrum',None):
      if obj is not None and hasattr(obj,tag):
        return getattr(obj,tag);
    return None;
    

  def source_list (self,ns,max_sources=None,**kw):
    """Reads LSM and returns a list of Meow objects.
    ns is node scope in which they will be created.
    Keyword arguments may be used to indicate which of the source attributes are to be
    created as Parms, use e.g. I=Meow.Parm(tags="flux") for this.
    The use_parms option may override this.
    """;
    if self.filename is None:
      return [];
    # load the sky model
    if self.lsm is None:
      self.lsm = Tigger.load(self.filename);

    # sort by brightness
    sources = sorted(self.lsm.sources,lambda a,b:cmp(b.brightness(),a.brightness()));

    # extract subset, if specified
    sources = SourceSubsetSelector.filter_subset(self.lsm_subset,sources,self._getTagValue);
    # get nulls subset
    if self.null_subset:
      nulls = set([src.name for src in SourceSubsetSelector.filter_subset(self.null_subset,sources)]);
    else:
      nulls = set();
    parm = Meow.Parm(tags="source solvable");
    # make copy of kw dict to be used for sources not in solvable set
    kw_nonsolve = dict(kw);
    # and update kw dict to be used for sources in solvable set
    # this will be a dict of lists of solvable subgroups
    parms = [];
    subgroups = {};
    if self.solvable_sources:
      subgroup_order = [];
      for sgname in _SubgroupOrder:
        if getattr(self,'solve_%s'%sgname):
          sg = subgroups[sgname] = [];
          subgroup_order.append(sgname);

    # make Meow list
    source_model = []

    for src in sources:
      is_null = src.name in nulls;
      # this will be True if this source has solvable parms
      solvable = self.solvable_sources and not is_null and ( not self.lsm_solvable_tag
                  or getattr(src,self.lsm_solvable_tag,False) );
      if solvable:
        # independent groups?
        if self.lsm_solve_group_tag:
          independent_sg = sgname = "%s:%s"%(self.lsm_solve_group_tag,getattr(src,self.lsm_solve_group_tag,"unknown"));
        else:
          independent_sg = "";
          sgname = 'source:%s'%src.name;
        if sgname in subgroups:
          sgsource = subgroups[sgname];
        else:
          sgsource = subgroups[sgname] = [];
          subgroup_order.append(sgname);
      # make dict of source parametrs: for each parameter we have a value,subgroup pair
      if is_null:
        attrs = dict(ra=src.pos.ra,dec=src.pos.dec,I=0,Q=None,U=None,V=None,RM=None,spi=None,freq0=None);
      else:
        attrs = dict(
          ra=     src.pos.ra,
          dec=    src.pos.dec,
          I=      src.flux.I,
          Q=      getattr(src.flux,'Q',None),
          U=      getattr(src.flux,'U',None),
          V=      getattr(src.flux,'V',None),
          RM=     getattr(src.flux,'rm',None),
          freq0=  getattr(src.flux,'freq0',None) or (src.spectrum and getattr(src.spectrum,'freq0',None)),
          spi=    src.spectrum and getattr(src.spectrum,'spi',None)
        );
      if not is_null and isinstance(src.shape,ModelClasses.Gaussian):
        attrs['lproj'] = src.shape.ex*math.sin(src.shape.pa);
        attrs['mproj'] = src.shape.ex*math.cos(src.shape.pa);
        attrs['ratio'] = src.shape.ey/src.shape.ex;
      # construct parms or constants for source attributes, depending on whether the source is solvable or not
      # If source is solvable and this particular attribute is solvable, replace
      # value in attrs dict with a Meq.Parm.
      if solvable:
        for parmname,value in attrs.items():
          sgname = _Subgroups.get(parmname,None);
          if sgname in subgroups:
            solvable = True;
            parm = attrs[parmname] = ns[src.name](parmname) << Meq.Parm(value or 0,
                                                                tags=["solvable",sgname],solve_group=independent_sg);
            subgroups[sgname].append(parm);
            sgsource.append(parm);
            parms.append(parm);

      # construct a direction
      direction = Meow.Direction(ns,src.name,attrs['ra'],attrs['dec'],static=not solvable or not self.solve_pos);

      # construct a point source or gaussian or FITS image, depending on source shape class
      if src.shape is None or is_null:
        msrc = Meow.PointSource(ns,name=src.name,
                I=attrs['I'],Q=attrs['Q'],U=attrs['U'],V=attrs['V'],
                direction=direction,
                spi=attrs['spi'],freq0=attrs['freq0'],RM=attrs['RM']);
      elif isinstance(src.shape,ModelClasses.Gaussian):
        msrc = Meow.GaussianSource(ns,name=src.name,
                I=attrs['I'],Q=attrs['Q'],U=attrs['U'],V=attrs['V'],
                direction=direction,
                spi=attrs['spi'],freq0=attrs['freq0'],
                lproj=attrs['lproj'],mproj=attrs['mproj'],ratio=attrs['ratio']);
        if solvable and 'shape' in subgroups:
          subgroups['pos'] += direction.get_solvables();
      elif isinstance(src.shape,ModelClasses.FITSImage):
        msrc = Meow.FITSImageComponent(ns,name=src.name,
                    filename=src.shape.filename,
                    direction=direction);
        msrc.set_options(fft_pad_factor=(src.shape.pad or 2));

      msrc.solvable = solvable;

      # copy standard attributes from sub-objects
      for subobj in src.flux,src.shape,src.spectrum:
        if subobj:
          for attr,val in src.flux.getAttributes():
            msrc.set_attr(attr,val);
      # copy all extra attrs from source object
      for attr,val in src.getExtraAttributes():
        msrc.set_attr(attr,val);

      # make sure Iapp exists (init with I if it doesn't)
      if msrc.get_attr('Iapp',None) is None:
        msrc.set_attr('Iapp',src.flux.I);

      source_model.append(msrc);

    # if any solvable parms were made, make a parmgroup and solve job for them
    if parms:
      if os.path.isdir(self.filename):
        table_name = os.path.join(self.filename,"sources.fmep");
      else:
        table_name = os.path.splitext(self.filename)[0]+".fmep";
      # make list of Subgroup objects for every non-empty subgroup
      sgs = [];
      for sgname in subgroup_order:
        sglist = subgroups.get(sgname,None);
        if sglist:
          sgs.append(Meow.ParmGroup.Subgroup(sgname,sglist));
      # make main parm group
      pg_src = Meow.ParmGroup.ParmGroup("source parameters",parms,
                  subgroups=sgs,
                  table_name=table_name,table_in_ms=False,bookmark=True);
      # now make a solvejobs for the source
      Meow.ParmGroup.SolveJob("cal_source","Solve for source parameters",pg_src);


    return source_model;




# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import time
import os.path
import sys
import traceback
from HTMLParser import HTMLParser

import Kittens.utils
_verbosity = Kittens.utils.verbosity(name="lsmhtml");
dprint = _verbosity.dprint;
dprintf = _verbosity.dprintf;

from Tigger.Models import ModelClasses
from Tigger.Models import SkyModel

DefaultExtension = "lsm.html";

def save (model,filename,sources=None,**kw):
  if sources is None:
    sources = model.sources;
  fobj = file(filename,'w');
  fobj.write("""<HTML><BODY mdltype=SkyModel>\n""");
  if model.name is not None:
    fobj.write(model.renderAttrMarkup('name',model.name,tags='TITLE',verbose="Sky model: "));
    fobj.write("\n");
  # write list of sources
  fobj.write("""<H1>Source list</H1>\n<TABLE BORDER=1 FRAME=box RULES=all CELLPADDING=5>\n""");
  for src in sources:
    fobj.write(src.renderMarkup(tags=["TR\n","TD"]));
    fobj.write("\n");
  fobj.write("""</TABLE>\n""");
  # plot styles
  if model.plotstyles is not None:
    fobj.write("""<H1>Plot styles</H1>\n<TABLE BORDER=1 FRAME=box RULES=all CELLPADDING=5>\n""");
    fobj.write(model.renderAttrMarkup('plotstyles',model.plotstyles,tags=['A','TR\n','TD'],verbose=""));
    fobj.write("""</TABLE>\n""");
  # other attributes
  fobj.write("\n");
  fobj.write("""<H1>Other properties</H1>\n""");
  if model.pbexp is not None:
    fobj.write("<P>");
    fobj.write(model.renderAttrMarkup('pbexp',model.pbexp,tags='A',verbose="Primary beam expression: "));
    fobj.write("</P>\n");
  if model.freq0 is not None:
    fobj.write("<P>");
    fobj.write(model.renderAttrMarkup('freq0',model.freq0,tags='A',verbose="Reference frequency, Hz: "));
    fobj.write("</P>\n");
  if model.ra0 is not None or model.dec0 is not None:
    fobj.write("<P>");
    fobj.write(model.renderAttrMarkup('ra0',model.ra0,tags='A',verbose="Field centre ra: "));
    fobj.write(model.renderAttrMarkup('dec0',model.dec0,tags='A',verbose="dec: "));
    fobj.write("</P>\n");
  for attr,value in model.getExtraAttributes():
    if attr not in ("pbexp","freq0","ra0","dec0"):
      fobj.write("<P>");
      fobj.write(model.renderAttrMarkup(attr,value,tags='A'));
      fobj.write("</P>\n");
  fobj.write("""</BODY></HTML>\n""");

def load (filename,**kw):
  parser = ModelIndexParser();
  parser.reset();
  for line in file(filename):
      parser.feed(line);
  parser.close();
  if not parser.toplevel_objects:
    raise RuntimeError,"failed to load sky model from file %s"%filename;
  return parser.toplevel_objects[0];

class ModelIndexParser (HTMLParser):
  def reset (self):
    HTMLParser.reset(self);
    self.objstack = [];
    self.tagstack = [];
    self.toplevel_objects = [];

  def end (self):
    dprintf(4,"end");

  def handle_starttag (self,tag,attrs):
    dprint(4,"start tag",tag,attrs);
    attrs = dict(attrs);
    # append tag to tag stack. Second element in tuple indicates whether
    # tag is associated with the start of an object definition
    self.tagstack.append([tag,None]);
    # see if attributes describe an LSM object
    # 'type' is an object class
    mdltype = attrs.get('mdltype');
    if not mdltype:
      return;
    # 'attr' is an attribute name. If this is set, then the object is an attribute
    # of the parent-level class
    mdlattr = attrs.get('mdlattr');
    # 'value' is a value. If this is set, then the object can be created from a string
    mdlval  = attrs.get('mdlval');
    dprintf(3,"model item type %s, attribute %s, inline value %s\n",mdltype,mdlattr,mdlval);
    if mdlattr and not self.objstack:
      dprintf(0,"WARNING: attribute %s at top level, ignoring\n",mdlattr);
      return;
    # Now look up the class in our globals, or in ModelClasses
    typeobj = ModelClasses.AtomicTypes.get(mdltype) or ModelClasses.__dict__.get(mdltype);
    if not callable(typeobj):
      dprintf(0,"WARNING: unknown object type %s, ignoring\n",mdltype);
      return;
    # see if object value is inlined
    if mdlval is not None:
      try:
        obj = typeobj(eval(mdlval));
      except:
        traceback.print_exc();
        dprintf(0,"WARNING: failed to create object of type %s from string value '%s', ignoring\n",mdltype,mdlval);
        return;
      self.add_object(mdlattr,obj);
    # else add object to stack and start accumulating attributes
    else:
      # change entry on tagstack to indicate that this tag started an object
      self.tagstack[-1][1] = len(self.objstack);
      # append object entry to stack -- we'll create the object when a corresponding end-tag
      # is encountered.
      self.objstack.append([mdlattr,typeobj,[],{}]);

  def handle_endtag (self,endtag):
    dprint(4,"end tag",endtag);
    # close all tags from top of stack, until we hit this one's start tag
    while self.tagstack:
      tag,nobj = self.tagstack.pop(-1);
      dprint(4,"closing tag",tag);
      # if tag corresponds to an object, create object
      if nobj is not None:
        self.close_stack_object();
      if tag == endtag:
        break;

  def add_object (self,attr,obj):
    """Adds object to model."""
    # if no object stack, then object is a top-level container
    if not self.objstack:
      if attr:
        dprintf(0,"WARNING: attribute %s at top level, ignoring\n",attr);
        return;
      self.toplevel_objects.append(obj);
    # else  add object as attribute or argument of top container in the stack
    else:
      if attr:
        self.objstack[-1][3][attr] = obj;
      else:
        self.objstack[-1][2].append(obj);

  def close_stack_object (self):
    """This function is called when an object from the top of the stack needs to be created.
    Stops accumulating attributes and calls the object constructor."""
    mdlattr,typeobj,args,kws = self.objstack.pop(-1);
    # create object
    try:
      if typeobj in (list,tuple):
        obj = typeobj(args);
      else:
        obj = typeobj(*args,**kws);
    except:
      traceback.print_exc();
      dprintf(0,"WARNING: failed to create object of type %s for attribute %s, ignoring\n",typeobj,mdlattr);
      return;
    # add the object to model
    self.add_object(mdlattr,obj);

import Tigger.Models.Formats
Tigger.Models.Formats.registerFormat("Tigger",load,"Tigger sky model",(".lsm.html",),export_func=save);



# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import math
import os.path
import numpy
import copy

from Tigger import startup_dprint
startup_dprint(1,"starting ModelClasses");

DEG = 180/math.pi;

AtomicTypes = dict(bool=bool,int=int,float=float,complex=complex,str=str,list=list,tuple=tuple,dict=dict,NoneType=lambda x:None);

class ModelItem (object):
  """ModelItem is a base class for all model items. ModelItem provides functions
  for saving, loading, and initializing items, using class attributes that describe the
  item's structure.
  A ModelItem has a number of named attributes (both mandatory and optional), which are
    sufficient to fully describe the item.
  A ModelItem is constructed by specifying its attribute values. Mandatory attributes are
    passed as positional arguments to the constructor, while optional attributes are passed
    as keyword arguments.
  'mandatory_attrs' is a class data member that provides a list of mandatory attributes.
  'optional_attrs' is a class data member that provides a dict of optional attributes and their
      default values (i.e. their value when missing). Subclasses are expected to redefine these
      attributes.
  """;

  # list of mandatory item attributes
  mandatory_attrs  = [];
  # dict of optional item attributes (key is name, value is default value)
  optional_attrs   = {};
  # True is arbitrary extra attributes are allowed
  allow_extra_attrs = False;
  # dict of rendertags for attributes. Default is to render ModelItems with the "A" tag,
  # and atomic attributes with the "TD" tag
  attr_rendertag   = {};
  # dict of verbosities for attributes. If an entry is present for a given attribute, then
  # the attribute's text representation will be rendered within its tags
  attr_verbose     = {};

  def __init__ (self,*args,**kws):
    """The default ModelItem constructor treats its positional arguments as a list of
    mandatory attributes, and its keyword arguments as optional attributes""";
    # check for argument errors
    if len(args) < len(self.mandatory_attrs):
      raise TypeError,"too few arguments in constructor of "+self.__class__.__name__;
    if len(args) > len(self.mandatory_attrs):
      raise TypeError,"too many arguments in constructor of "+self.__class__.__name__;
    # set mandatory attributes from argument list
    for attr,value in zip(self.mandatory_attrs,args):
      if not isinstance(value,AllowedTypesTuple):
        raise TypeError,"invalid type %s for attribute %s (class %s)"%(type(value).__name__,attr,self.__class__.__name__);
      setattr(self,attr,value);
    # set optional attributes from keywords
    for kw,default in self.optional_attrs.iteritems():
      value = kws.pop(kw,default);
      if not isinstance(value,AllowedTypesTuple):
        raise TypeError,"invalid type %s for attribute %s (class %s)"%(type(value).__name__,kw,self.__class__.__name__);
      setattr(self,kw,value);
    # set extra attributes, if any are left
    self._extra_attrs = set();
    if self.allow_extra_attrs:
      for kw,value in kws.iteritems():
        if not isinstance(value,AllowedTypesTuple):
          raise TypeError,"invalid type %s for attribute %s (class %s)"%(type(value).__name__,kw,self.__class__.__name__);
        self.setAttribute(kw,value);
    elif kws:
        raise TypeError,"unknown parameters %s in constructor of %s"%(','.join(kws.keys()),self.__class__.__name__);
    # other init
    self._signaller = None;
    self._connections = set();

  def enableSignals (self):
    """Enables Qt signals for this object.""";
    import PyQt4.Qt;
    self._signaller = PyQt4.Qt.QObject();

  def signalsEnabled (self):
    return bool(self._signaller);

  def connect (self,signal_name,receiver,reconnect=False):
    """Connects SIGNAL from object to specified receiver slot. If reconnect is True, allows duplicate connections.""";
    if not self._signaller:
      raise RuntimeError,"ModelItem.connect() called before enableSignals()";
    import PyQt4.Qt;
    if reconnect or (signal_name,receiver) not in self._connections:
      self._connections.add((signal_name,receiver));
      PyQt4.Qt.QObject.connect(self._signaller,PyQt4.Qt.SIGNAL(signal_name),receiver);

  def emit (self,signal_name,*args):
    """Emits named SIGNAL from this object .""";
    if not self._signaller:
      raise RuntimeError,"ModelItem.emit() called before enableSignals()";
    import PyQt4.Qt;
    self._signaller.emit(PyQt4.Qt.SIGNAL(signal_name),*args);

  def registerClass (classobj):
    if not isinstance(classobj,type):
      raise TypeError,"registering invalid class object: %s"%classobj;
    globals()[classobj.__name__] = classobj;
    AllowedTypes[classobj.__name__] = classobj;
    AllowedTypesTuple = tuple(AllowedTypes.itervalues());
  registerClass = classmethod(registerClass);

  def setAttribute (self,attr,value):
    if attr not in self.mandatory_attrs and attr not in self.optional_attrs:
      self._extra_attrs.add(attr);
    setattr(self,attr,value);

  def removeAttribute (self,attr):
    if hasattr(self,attr):
      delattr(self,attr);
    self._extra_attrs.discard(attr);

  def getExtraAttributes (self):
    """Returns list of extra attributes, as (attr,value) tuples""";
    return  [ (attr,getattr(self,attr)) for attr in self._extra_attrs ];

  def getAttributes (self):
    """Returns list of all attributes (mandatory+optional+extra), as (attr,value) tuples""";
    attrs = [ (attr,getattr(self,attr)) for attr in self.mandatory_attrs ];
    for attr,default in self.optional_attrs.iteritems():
      val = getattr(self,attr,default);
      if val != default:
        attrs.append((attr,val));
    attrs += [ (attr,getattr(self,attr)) for attr in self._extra_attrs ];
    return attrs;

  def __copy__ (self):
    """Returns copy of object. Copies all attributes.""";
    attrs = self.optional_attrs.copy();
    attrs.update(self.getExtraAttributes());
    return self.__class__( *[ getattr(self,attr) for attr in self.mandatory_attrs],**attrs);
  
  def __deepcopy__ (self,memodict):
    """Returns copy of object. Copies all attributes.""";
    attrs = self.optional_attrs.copy();
    attrs.update(self.getExtraAttributes());
    attrs = copy.deepcopy(attrs,memodict);
    return self.__class__( *[ copy.deepcopy(getattr(self,attr),memodict) for attr in self.mandatory_attrs],**attrs);

  def copy (self,deep=True):
    if deep:
      return copy.deepcopy(self);
    else:
      return __copy__(self);

  def strAttributes (self,sep=",",label=True,
                                float_format="%.2g",complex_format="%.2g%+.2gj"):
    """Renders attributes as string. Child classes may redefine this to make a better string representation.
    If label=True, uses "attrname=value", else uses "value".
    'sep' specifies a separator.
    """;
    fields = [];
    for attr,val in self.getAttributes():
      ss = (label and "%s="%attr) or "";
      if isinstance(val,(float,int)):
        ss += float_format%val;
      elif isinstance(val,complex):
        ss += complex_format%val;
      else:
        ss += str(val);
      fields.append(ss);
    return sep.join(fields);

  def strDesc (self,**kw):
    """Returns string describing the object, used in GUIs and such. Default implementation calls strAttributes()."""
    return strAttributes(**kw);

  def _resolveTags (self,tags,attr=None):
    """helper function called from renderMarkup() and renderAttrMarkup() below to
    figure out which HTML tags to enclose a value in. Return value is tuple of (tag,endtag,rem_tags), where
    tag is the HTML tag to use (or None for default, usually "A"), endtag is the closing tag (including <> and whitespace, if any),
    and rem_tags is to be passed to child items' resolveMarkup() """;
    # figure out enclosing tag
    if not tags:
      tag,tags = None,None;  # use default
    elif isinstance(tags,str):
      tag,tags = tags,None;           # one tag supplied, use that here and use defaults for sub-items
    elif isinstance(tags,(list,tuple)):
      tag,tags = tags[0],tags[1:];   # stack of tags supplied: use first here, pass rest to sub-items
    else:
      raise ValueError,"invalid 'tags' parameter of type "+str(type(tags));
    # if tag is None, use default
    tag = tag or self.attr_rendertag.get(attr,None) or "A";
    if tag.endswith('\n'):
      tag = tag[:-1];
      endtag = "</%s>\n"%tag;
    else:
      endtag = "</%s> "%tag;
    return tag,endtag,tags;

  def renderMarkup (self,tags=None,attrname=None):
    """Makes a markup string corresponding to the model item.
    'tags' is the HTML tag to use.
    If 'verbose' is not None, a text representation of the item (using str()) will be included
    as HTML text between the opening and closing tags.
    """;
    tag,endtag,tags = self._resolveTags(tags,attrname);
    # opening tag
    markup = "<%s mdltype=%s "%(tag,type(self).__name__);
    if attrname is not None:
      markup += "mdlattr=\"%s\" "%attrname;
    markup +=">";
    # render attrname as comment
    if attrname:
      if tag == "TR":
        markup += "<TD bgcolor=yellow>%s</TD>"%attrname;
      else:
        markup += "<A>%s:</A> "%attrname;
    # write mandatory attributes
    for attr in self.mandatory_attrs:
      markup += self.renderAttrMarkup(attr,getattr(self,attr),tags=tags,mandatory=True);
    # write optional attributes only wheh non-default
    for attr,default in sorted(self.optional_attrs.iteritems()):
      val = getattr(self,attr,default);
      if val != default:
        markup += self.renderAttrMarkup(attr,val,tags=tags);
    # write extra attributes
    for attr in self._extra_attrs:
      markup += self.renderAttrMarkup(attr,getattr(self,attr),tags=tags);
    # closing tag
    markup += endtag;
    return markup;
    
  numpy_int_types = tuple([
      getattr(numpy,"%s%d"%(t,d)) for t in "int","uint" for d in 8,16,32,64 
      if hasattr(numpy,"%s%d"%(t,d))
    ]);
  numpy_float_types = tuple([
      getattr(numpy,"float%d"%d) for d in 32,64,96,128 
      if hasattr(numpy,"float%d"%d)
    ]);

  def renderAttrMarkup (self,attr,value,tags=None,verbose=None,mandatory=False):
    # render ModelItems recursively via renderMarkup() above
    if isinstance(value,ModelItem):
      return value.renderMarkup(tags,attrname=(not mandatory and attr) or None);
    # figure out enclosing tags
    tag,endtag,tags = self._resolveTags(tags,attr);
    # convert numpy types to float or complexes
    if isinstance(value,self.numpy_int_types):
      value = int(value);
    elif isinstance(value,self.numpy_float_types):
      value = float(value);
    elif numpy.iscomplexobj(value):
      value = complex(value);
    # render opening tags
    markup = "<%s mdltype=%s "%(tag,type(value).__name__);
    if not mandatory:
      markup += "mdlattr=\"%s\" "%attr;
    # if rendering table row, use TD to render comments
    if verbose is None:
      verbose = attr; # and self.attr_verbose.get(attr);
    if tag == "TR":
      comment = "<TD bgcolor=yellow>%s</TD>";
    else:
      comment = "<A>%s</A> ";
    # render lists or tuples iteratively
    if isinstance(value,(list,tuple)):
      markup += ">";
      if verbose:
        markup += comment%(verbose+":");
      for i,item in enumerate(value):
        markup += self.renderAttrMarkup(str(i),item,mandatory=True,tags=tags);
    # render dicts iteratively
    elif isinstance(value,dict):
      markup += ">";
      if verbose:
        markup += comment%(verbose+":");
      for key,item in sorted(value.iteritems()):
        markup += self.renderAttrMarkup(key,item,tags=tags);
    # render everything else inline
    else:
      if isinstance(value,str):
        markup += "mdlval=\"'%s'\">"%value.replace("\"","\\\"").replace("'","\\'");
      else:
        markup += "mdlval=\"%s\">"%repr(value);
      if verbose is attr:
        markup += comment%':'.join((attr,str(value)));
      else:
        markup += comment%''.join((verbose,str(value)));
    markup += endtag;
    return markup;

def _deg_to_dms (x,prec=0.01):
  """Converts x (in degrees) into d,m,s tuple, where d and m are ints.
  prec gives the precision, in arcseconds."""
  mins,secs = divmod(round(x*3600/prec)*prec,60);
  mins = int(mins);
  degs,mins = divmod(mins,60);
  return degs,mins,secs;

class Position (ModelItem):
  mandatory_attrs  = [ "ra","dec" ];
  optional_attrs = dict(ra_err=None,dec_err=None);

  @staticmethod
  def ra_hms_static (rad,scale=12,prec=0.01):
    """Returns RA as tuple of (h,m,s)""";
    # convert negative values
    while rad < 0:
        rad += 2*math.pi;
    # convert to hours
    rad *= scale/math.pi;
    return  _deg_to_dms(rad,prec);

  def ra_hms (self,prec=0.01):
    return self.ra_hms_static(self.ra,scale=12,prec=prec);

  def ra_dms (self,prec=0.01):
    return self.ra_hms_static(self.ra,scale=180,prec=prec);

  @staticmethod
  def dec_dms_static (rad,prec=0.01):
    return Position.dec_sdms_static(rad,prec)[1:];
  
  @staticmethod
  def dec_sdms_static (rad,prec=0.01):
    """Returns Dec as tuple of (sign,d,m,s). Sign is "+" or "-".""";
    sign = "-" if rad<0 else "+";
    d,m,s = _deg_to_dms(abs(rad)*DEG,prec);
    return (sign,d,m,s);

  def dec_sdms (self,prec=0.01):
    return self.dec_sdms_static(self.dec,prec);

class Flux (ModelItem):
  mandatory_attrs  = [ "I" ];
  optional_attrs = dict(I_err=None);
  def rescale (self,scale):
    self.I *= scale;

class Polarization (Flux):
  mandatory_attrs  = Flux.mandatory_attrs + [ "Q","U","V" ];
  optional_attrs = dict(I_err=None,Q_err=None,U_err=None,V_err=None);
  def rescale (self,scale):
    for stokes in "IQUV":
      setattr(self,stokes,getattr(self,stokes)*scale);

class PolarizationWithRM (Polarization):
  mandatory_attrs = Polarization.mandatory_attrs + [ "rm","freq0" ];
  optional_attrs = dict(Polarization.optional_attrs,rm_err=None)

class Spectrum (ModelItem):
  """The Spectrum class is an abstract representation of spectral information. The base implementation corresponds
  to a flat spectrum.
  """;
  def normalized_intensity (self,freq):
    """Returns the normalized intensity for a given frequency,normalized to unity at the reference frequency (if any)"""
    return 1;

class SpectralIndex (Spectrum):
  mandatory_attrs  = [ "spi","freq0" ];
  optional_attrs = dict(spi_err=None);
  def normalized_intensity (self,freq):
    """Returns the normalized intensity for a given frequency, normalized to unity at the reference frequency (if any)"""
    if isinstance(self.spi,(list,tuple)):
      spi = self.spi[0];
      logfreq = numpy.log(freq/self.freq0);
      for i,x in enumerate(self.spi[1:]):
        spi = spi + x*(logfreq**(i+1));
    else:
      spi = self.spi;
    return (freq/self.freq0)**spi;

class Shape (ModelItem):
  """Abstract base class for a source's brightness distribution.
  The ex/ey/pa attributes give the overall shape of the source."""
  mandatory_attrs  = [ "ex","ey","pa" ];
  optional_attrs = dict(ex_err=None,ey_err=None,pa_err=None);
  def getShape (self):
    return self.ex,self.ey,self.pa
  def getShapeErr (self):
    err = [ getattr(self,a+'_err',None) for a in self.mandatory_attrs ] 
    if all([ a is None for a in err ]):
      return None
    return tuple(err)

class Gaussian (Shape):
  typecode = "Gau";
  def strDesc (self,delimiters=('"',"x","@","deg"),**kw):
    return """%.2g%s%s%.2g%s%s%d%s"""%(self.ex*DEG*3600,delimiters[0],delimiters[1],self.ey*DEG*3600,delimiters[0],
                                       delimiters[2],round(self.pa*DEG),delimiters[3]);
  def strDescErr (self,delimiters=('"',"x","@","deg"),**kw):
    err = self.getShapeErr();
    return err and """%.2g%s%s%.2g%s%s%d%s"""%(err[0]*DEG*3600,delimiters[0],delimiters[1],err[1]*DEG*3600,delimiters[0],
                                       delimiters[2],round(err[2]*DEG),delimiters[3]);


class FITSImage (Shape):
  typecode = "FITS";
  mandatory_attrs  = Shape.mandatory_attrs + [ "filename","nx","ny" ];
  optional_attrs = dict(pad=2);
  def strDesc (self,**kw):
    return """%s %dx%d"""%(os.path.basename(self.filename),self.nx,self.ny);

startup_dprint(1,"end of class defs");

# populate dict of AllowedTypes with all classes defined so far
globs = list(globals().iteritems());

AllowedTypes = dict(AtomicTypes.iteritems());
AllowedTypes['NoneType'] = type(None);  # this must be a type, otherwise isinstance() doesn't work
for name,val in globs:
  if isinstance(val,type):
    AllowedTypes[name] = val;
AllowedTypesTuple = tuple(AllowedTypes.itervalues());

startup_dprint(1,"end of ModelClasses");



# -*- coding: utf-8 -*-
#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import Kittens.utils
import os.path
import sys
import traceback

_verbosity = Kittens.utils.verbosity(name="lsmformats");
dprint = _verbosity.dprint;
dprintf = _verbosity.dprintf;

Formats = {};
_FormatList = [];
_FormatsInitialized = False;

def _initFormats ():
  """Initializes all known formats by importing their modules""";
  global _FormatsInitialized;
  if not _FormatsInitialized:
    for format in [ "ModelHTML","ASCII","BBS","NEWSTAR","AIPSCC","AIPSCCFITS","PyBDSMGaul" ]:
      try:
        __import__(format,globals(),locals());
      except:
        traceback.print_exc();
        print "Error loading support for format '%s', see above. Format will not be available."%format;
    _FormatsInitialized = True;

def registerFormat (name,import_func,doc,extensions,export_func=None):
  """Registers an external format, with an import function""";
  global Formats;
  Formats[name] = (import_func,export_func,doc,extensions);
  _FormatList.append(name);

def getFormat (name):
  """Gets file format by name. Returns name,import_func,export_func,docstring if found, None,None,None,None otherwise.""";
  _initFormats();
  if name not in Formats:
    return None,None,None,None;
  import_func,export_func,doc,extensions = Formats[name];
  return name,import_func,export_func,doc;

def getFormatExtensions (name):
  """Gets file format by name. Returns name,import_func,export_func,docstring if found, None,None,None,None otherwise.""";
  _initFormats();
  if name not in Formats:
    return None;
  import_func,export_func,doc,extensions = Formats[name];
  return extensions;
  
def determineFormat (filename):
  """Tries to determine file format by filename. Returns name,import_func,export_func,docstring if found, None,None,None,None otherwise.""";
  _initFormats();
  for name,(import_func,export_func,doc,extensions) in Formats.iteritems():
    for ext in extensions:
      if filename.endswith(ext):
        return name,import_func,export_func,doc;
  return None,None,None,None;
  
def listFormats ():
  _initFormats();
  return _FormatList;

def listFormatsFull ():
  _initFormats();
  return [ (name,Formats[name]) for name in _FormatList ];

def resolveFormat (filename,format):
  """Helper function, resolves format/filename arguments to a format tuple""";
  _initFormats();
  if format:
    name,import_func,export_func,doc = getFormat(format);
    if not import_func:
      raise TypeError("Unknown model format '%s'"%format);
  else:
    name,import_func,export_func,doc = determineFormat(filename);
    if not import_func:
      raise TypeError("Cannot determine model format from filename '%s'"%filename);
  return name,import_func,export_func,doc;
  
# provide some convenience methods

def load (filename,format=None,verbose=True):
  """Loads a sky model."""
  name,import_func,export_func,doc = resolveFormat(filename,format);
  if not import_func:
    raise TypeError("Unknown model format '%s'"%format);
  if verbose:
    print "Loading %s: %s"%(filename,doc);
  return import_func(filename);

def save (model,filename,format=None,verbose=True):
  """Saves a sky model."""
  name,import_func,export_func,doc = resolveFormat(filename,format);
  if verbose:
    print "Saving %s: %s"%(filename,doc);
  return export_func(model,filename);



# -*- coding: utf-8 -*-
#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import Tigger
from Tigger import startup_dprint
startup_dprint(1,"start of Coordinates");

import sys
import math
import numpy
from numpy import sin,cos,arcsin,arccos;
startup_dprint(1,"imported numpy");


import Kittens.utils
pyfits = Kittens.utils.import_pyfits();
startup_dprint(1,"imported pyfits");

DEG = math.pi/180;

startup_dprint(1,"importing WCS");

# If we're being imported outside the main app (e.g. a script is trying to read a Tigger model,
# whether TDL or otherwise), then pylab may be needed by that script for decent God-fearing
# purposes. Since WCS is going to pull it in anyway, we try to import it here, and if that
# fails, replace it by dummies.
if not Tigger.matplotlib_nuked:
  try:
    import pylab;
  except:
    Tigger.nuke_matplotlib();

# some locales cause WCS to complain that "." is not the decimal separator, so reset it to "C"
import locale
locale.setlocale(locale.LC_NUMERIC, 'C')
      

try:
  from astLib.astWCS import WCS
  import PyWCSTools.wcs
except ImportError:
  print "Failed to import the astLib.astWCS and/or PyWCSTools module. Please install the astLib package (http://astlib.sourceforge.net/)."
  raise;

startup_dprint(1,"imported WCS");

def angular_dist_pos_angle (ra1,dec1,ra2,dec2):
  """Computes the angular distance between the two points on a sphere, and
  the position angle (North through East) of the direction from 1 to 2.""";
  # I lifted this somewhere
  sind1,sind2 = sin(dec1),sin(dec2);
  cosd1,cosd2 = cos(dec1),cos(dec2);
  cosra,sinra = cos(ra1-ra2),sin(ra1-ra2);

  adist = numpy.arccos(min(sind1*sind2 + cosd1*cosd2*cosra,1));
  pa = numpy.arctan2(-cosd2*sinra,-cosd2*sind1*cosra+sind2*cosd1);
  return adist,pa;

def angular_dist_pos_angle2 (ra1,dec1,ra2,dec2):
  """Computes the angular distance between the two points on a sphere, and
  the position angle (North through East) of the direction from 1 to 2.""";
  # I re-derived this from Euler angles, but it seems to be identical to the above
  ra = ra2 - ra1;
  sind0,sind,cosd0,cosd = sin(dec1),sin(dec2),cos(dec1),cos(dec2);
  sina,cosa = sin(ra)*cosd,cos(ra)*cosd;
  x = cosa*sind0 - sind*cosd0;
  y = sina;
  z = cosa*cosd0 + sind*sind0;
  print x,y,z;
  PA = numpy.arctan2(y,-x);
  R = numpy.arccos(z);

  return R,PA;

def angular_dist_pos_angle2 (ra1,dec1,ra2,dec2):
  """Computes the angular distance between the two points on a sphere, and
  the position angle (North through East) of the direction from 1 to 2.""";
  # I re-derived this from Euler angles, but it seems to be identical to the above
  ra = ra2 - ra1;
  sind0,sind,cosd0,cosd = sin(dec1),sin(dec2),cos(dec1),cos(dec2);
  sina,cosa = sin(ra)*cosd,cos(ra)*cosd;
  x = cosa*sind0 - sind*cosd0;
  y = sina;
  z = cosa*cosd0 + sind*sind0;
  print x,y,z;
  PA = numpy.arctan2(y,-x);
  R = numpy.arccos(z);
  return R,PA;



def _deg_to_dms (x,prec=0.01):
  """Converts x (in degrees) into d,m,s tuple, where d and m are ints.
  prec gives the precision, in arcseconds."""
  mins,secs = divmod(round(x*3600/prec)*prec,60);
  mins = int(mins);
  degs,mins = divmod(mins,60);
  return degs,mins,secs;

def ra_hms (rad,scale=12,prec=0.01):
  """Returns RA as tuple of (h,m,s)""";
  # convert negative values
  while rad < 0:
      rad += 2*math.pi;
  # convert to hours
  rad *= scale/math.pi;
  return  _deg_to_dms(rad,prec);

def dec_dms (rad,prec=0.01):
  return dec_sdms(rad,prec)[1:];

def dec_sdms (rad,prec=0.01):
  """Returns Dec as tuple of (sign,d,m,s). Sign is "+" or "-".""";
  sign = "-" if rad<0 else "+";
  d,m,s = _deg_to_dms(abs(rad)/DEG,prec);
  return (sign,d,m,s);

def ra_hms_string (rad):
  return "%dh%02dm%05.2fs"%ra_hms(rad);

def dec_sdms_string (rad):
  return "%s%dd%02dm%05.2fs"%dec_sdms(rad);

def radec_string (ra,dec):
  return "%s %s"%(ra_hms_string(ra),dec_sdms_string(dec));

class _Projector (object):
    """This is an abstract base class for all projection classes below. A projection class can be used to create projector objects for
    conversion between world (ra,dec) and projected (l,m) coordinates.

    * A projector is instantiated as proj = Proj(ra0,dec0)      # ra0,dec0 is projection centre
    * converts ra,dec->l,m as
          l,m = proj.lm(ra,dec)
    * converts l,m->ra,dec as
          ra,dec = proj.radec(l,m)
    * converts angular offsets (from 0,0 point) into l,m:
          l,m = proj.offset(dra,ddec);

    Alternativelty, there are class methods which do not require one to instantiate a projector object:

    * Proj.radec_lm(ra,dec,ra0,dec0)
    * Proj.lm_radec(l,m,ra0,dec0)
    * Proj.offset_lm(dra,ddec,ra0,dec0)
    """
    def __init__ (self,ra0,dec0,has_projection=False):
      self.ra0,self.dec0,self.sin_dec0,self.cos_dec0 = ra0,dec0,sin(dec0),cos(dec0);
      self._has_projection = has_projection;

    def has_projection (self):
      return bool(self._has_projection);

    def __eq__ (self,other):
      """By default, two projections are the same if their classes match, and their ra0/dec0 match."""
      return type(self) is type(other) and self.ra0 == other.ra0 and self.dec0 == other.dec0;

    def __ne__ (self,other):
      return not self == other;

    @classmethod
    def radec_lm (cls,ra,dec,ra0,dec0):
      return cls(ra0,dec0).lm(ra,dec);

    @classmethod
    def lm_radec (cls,l,m,ra0,dec0):
      return cls(ra0,dec0).radec(l,m);

    @classmethod
    def offset_lm (cls,dra,ddec,ra0,dec0):
      return cls(ra0,dec0).offset(dra,ddec);

    def lm (self,ra,dec):
      raise TypeError,"lm() not yet implemented in projection %s"%type(self).__name__;

    def offset (self,dra,ddec):
      raise TypeError,"offset() not yet implemented in projection %s"%type(self).__name__;

    def radec (self,l,m):
      raise TypeError,"radec() not yet implemented in projection %s"%type(self).__name__;

class Projection (object):
  """Projection is a container for the different projection classes.
  Each Projection class can be used to create a projection object: proj = Proj(ra0,dec0), with lm(ra,dec) and radec(l,m) methods.
  """;

  class FITSWCSpix (_Projector):
    """FITS WCS projection, as determined by a FITS header. lm is in pixels (0-based)."""
    def __init__ (self,header):
      """Constructor. Create from filename (treated as FITS file), or a FITS header object""";
      # attach to FITS file or header
      if isinstance(header,str):
        header = pyfits.open(header)[0].header;
      else:
        self.wcs = WCS(header,mode="pyfits");
      try:
        ra0,dec0 = self.wcs.getCentreWCSCoords();
        self.xpix0,self.ypix0 = self.wcs.wcs2pix(*self.wcs.getCentreWCSCoords());
        self.xscale = self.wcs.getXPixelSizeDeg()*DEG;
        self.yscale = self.wcs.getYPixelSizeDeg()*DEG;
        has_projection = True;
      except:
        print "No WCS in FITS file, falling back to pixel coordinates.";
        ra0 = dec0 = self.xpix0 = self.ypix0 = 0;
        self.xscale = self.yscale = DEG/3600;
        has_projection = False;
      _Projector.__init__(self,ra0*DEG,dec0*DEG,has_projection=has_projection);

    def lm (self,ra,dec):
      if not self.has_projection():
        return numpy.sin(ra)/self.xscale,numpy.sin(dec)/self.yscale;
      if numpy.isscalar(ra) and numpy.isscalar(dec):
        if ra - self.ra0 > math.pi:
          ra -= 2*math.pi;
        if ra - self.ra0 < -math.pi:
          ra += 2*math.pi;
        return self.wcs.wcs2pix(ra/DEG,dec/DEG);
      else:
        if numpy.isscalar(ra):
          ra = numpy.array(ra);
        ra[ra - self.ra0 > math.pi] -= 2*math.pi;
        ra[ra - self.ra0 < -math.pi] += 2*math.pi;
        ## when fed in arrays of ra/dec, wcs.wcs2pix will return a nested list of
        ## [[l1,m1],[l2,m2],,...]. Convert this to an array and extract columns.
        lm = numpy.array(self.wcs.wcs2pix(ra/DEG,dec/DEG));
        return lm[...,0],lm[...,1];

    def radec (self,l,m):
      if not self.has_projection():
        return numpy.arcsin(l*self.xscale),numpy.arcsin(m*self.yscale);
      if numpy.isscalar(l) and numpy.isscalar(m):
        ra,dec = self.wcs.pix2wcs(l,m);
      else:
## this is slow as molasses because of the way astLib.WCS implements the loop. ~120 seconds for 4M pixels
        ## when fed in arrays of ra/dec, wcs.wcs2pix will return a nested list of
        ## [[l1,m1],[l2,m2],,...]. Convert this to an array and extract columns.
#        radec = numpy.array(self.wcs.pix2wcs(l,m));
#        ra = radec[...,0];
#        dec = radec[...,1];
### try a faster implementation -- oh well, only a bit faster, ~95 seconds for the same
### can also replace list comprehension with map(), but that doesn't improve things.
### Note also that the final array constructor takes ~10 secs!
        radec = numpy.array([ PyWCSTools.wcs.pix2wcs(self.wcs.WCSStructure,x,y) for x,y in zip(l+1,m+1) ]);
        ra = radec[...,0];
        dec = radec[...,1];
      return ra*DEG,dec*DEG;


    def offset (self,dra,ddec):
      return self.xpix0 - dra/self.xscale,self.ypix0 + ddec/self.xscale;

    def __eq__ (self,other):
      """By default, two projections are the same if their classes match, and their ra0/dec0 match."""
      return type(self) is type(other) and (self.ra0,self.dec0,self.xpix0,self.ypix0,self.xscale,self.yscale)  == (other.ra0,other.dec0,other.xpix0,other.ypix0,other.xscale,other.yscale);

  class FITSWCS (FITSWCSpix):
    """FITS WCS projection, as determined by a FITS header. lm is renormalized to radians, l is reversed, 0,0 is at reference pixel."""
    def __init__ (self,header):
      """Constructor. Create from filename (treated as FITS file), or a FITS header object""";
      Projection.FITSWCSpix.__init__(self,header);

    def lm (self,ra,dec):
      if not self.has_projection():
        return -numpy.sin(ra)/self.xscale,numpy.sin(dec)/self.yscale;
      if numpy.isscalar(ra) and numpy.isscalar(dec):
        if ra - self.ra0 > math.pi:
          ra -= 2*math.pi;
        if ra - self.ra0 < -math.pi:
          ra += 2*math.pi;
        l,m = self.wcs.wcs2pix(ra/DEG,dec/DEG);
      else:
        if numpy.isscalar(ra):
          ra = numpy.array(ra);
        ra[ra - self.ra0 > math.pi] -= 2*math.pi;
        ra[ra - self.ra0 < -math.pi] += 2*math.pi;
        lm = numpy.array(self.wcs.wcs2pix(ra/DEG,dec/DEG));
        l,m = lm[...,0],lm[...,1];
      l = (self.xpix0-l)*self.xscale;
      m = (m-self.ypix0)*self.yscale;
      return l,m;

    def radec (self,l,m):
      if not self.has_projection():
        return numpy.arcsin(-l),numpy.arcsin(m);
      if numpy.isscalar(l) and numpy.isscalar(m):
        ra,dec = self.wcs.pix2wcs(self.xpix0-l/self.xscale,self.ypix0+m/self.yscale);
      else:
        radec = numpy.array(self.wcs.pix2wcs(self.xpix0-l/self.xscale,self.ypix0+m/self.yscale));
        ra = radec[...,0];
        dec = radec[...,1];
      return ra*DEG,dec*DEG;

    def offset (self,dra,ddec):
      return dra,ddec;

  @staticmethod
  def SinWCS (ra0,dec0):
    hdu = pyfits.PrimaryHDU();
    hdu.header.set('NAXIS',2);
    hdu.header.set('NAXIS1',3);
    hdu.header.set('NAXIS2',3);
    hdu.header.set('CTYPE1','RA---SIN');
    hdu.header.set('CDELT1',-1./60);
    hdu.header.set('CRPIX1',2);
    hdu.header.set('CRVAL1',ra0/DEG);
    hdu.header.set('CUNIT1','deg     ');
    hdu.header.set('CTYPE2','DEC--SIN');
    hdu.header.set('CDELT2',1./60);
    hdu.header.set('CRPIX2',2);
    hdu.header.set('CRVAL2',dec0/DEG);
    hdu.header.set('CUNIT2','deg     ');
    return Projection.FITSWCS(hdu.header);



# -*- coding: utf-8 -*-
#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys

from Tigger.Models.Formats import load, save, listFormats
import Kittens.config

__version__ = "1.4.2"

release_string = __version__
svn_revision_string = __version__
svn_revision_html = __version__

matplotlib_nuked = False

startup_dprint = startup_dprintf = lambda *dum:None
_verbosity = Kittens.utils.verbosity(name="tigger")
dprint = _verbosity.dprint
dprintf = _verbosity.dprintf

def import_pyfits ():
  """Helper function to import pyfits and return it. Provides a workaround for
  pyfits-2.3, which is actually arrogant enough (fuck you with a bargepole, pyfits!) 
  to replace the standard warnings.formatwarning function with its own BROKEN version, 
  thus breaking all other code that uses the warnings module."""
  if 'pyfits' not in sys.modules:
    import pyfits
    import warnings
    if getattr(pyfits,'formatwarning',None) is warnings.formatwarning:
      def why_is_pyfits_overriding_warnings_formatwarning_with_a_broken_one_damn_you_pyfits (message,category,  filename,lineno,line=None):
        return str(message)+'\n'
      warnings.formatwarning = why_is_pyfits_overriding_warnings_formatwarning_with_a_broken_one_damn_you_pyfits
    if getattr(pyfits,'showwarning',None) is warnings.showwarning:
      def showwarning_damn_you_pyfits_damn_you_sincerely (message,category,filename,lineno,file=None,line=None):
        pyfits.showwarning(message,category,filename,lineno,file=file)
      warnings.showwarning = showwarning_damn_you_pyfits_damn_you_sincerely
  return pyfits


def nuke_matplotlib ():
  """Some people think nothing of importing matplotlib at every opportunity, with no regard
  to consequences. Tragically, some of these people also write Python code, and some of them
  are responsible for astLib. Seriously man, if I just want to pull in WCS support, why the fuck
  do I need the monstrous entirety of matplotlib to come along with it, especially since it
  kills things like Qt outright?
  This function prevents such perversitities from happening, by inserting dummy modules
  into the sys.modules dict. Call nuke_matplotlib() once, and all further attempts to
  import matplotlib by any other code will be cheerfully ignored.
  """
  global matplotlib_nuked
  if 'pylab' not in sys.modules:
    # replace the modules referenced by astLib by dummy_module objects, which return a dummy callable for every attribute
    class dummy_module (object):
      def __getattr__ (self,name):
        return 'nowhere' if name == '__file__' else (lambda *args,**kw:True)
    sys.modules['pylab'] = sys.modules['matplotlib'] = sys.modules['matplotlib.patches'] = dummy_module()
    matplotlib_nuked = True



# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-

#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import Kittens.utils
pyfits = Kittens.utils.import_pyfits();
import math
import numpy

from Tigger.Coordinates import Projection
import FITSHeaders
from scipy.ndimage.filters import convolve
from scipy.ndimage.interpolation import map_coordinates
import astLib.astWCS

# init debug printing
import Kittens.utils
_verbosity = Kittens.utils.verbosity(name="imaging");
dprint = _verbosity.dprint;
dprintf = _verbosity.dprintf;

# conversion factors from radians
DEG = 180/math.pi;
ARCMIN = DEG*60;
ARCSEC = ARCMIN*60;
FWHM = math.sqrt(math.log(256));  # which is 2.3548;

def fitPsf (filename,cropsize=None):
  """Fits a Gaussian PSF to the FITS file given by 'filename'.
  If cropsize is specified, crops the central cropsize X cropsize pixels before fitting.
  Else determines cropsize by looking for the first negative sidelobe from the centre outwards.
  Returns maj_sigma,min_sigma,pa_NE (in radians)
  """;
  # read PSF from file
  psf = pyfits.open(filename)[0];
  hdr = psf.header;
  psf = psf.data;
  dprintf(2,"Read PSF of shape %s from file %s\n",psf.shape,filename);
  # remove stokes and freq axes
  if len(psf.shape) == 4:
    psf = psf[0,0,:,:];
  elif len(psf.shape) == 3:
    psf = psf[0,:,:];
  else:
    raise RuntimeError,"illegal PSF shape %s"+psf.shape;
  nx,ny = psf.shape;
  # crop the central region
  if cropsize:
    size = cropsize;
    psf = psf[(nx-size)//2:(nx+size)//2,(ny-size)//2:(ny+size)//2];
  # if size not specified, then auto-crop by looking for the first negative value starting from the center
  # this will break on very extended diagonal PSFs, but that's a pathological case
  else:
    ix = numpy.where(psf[:,ny//2]<0)[0];
    ix0 = max(ix[ix<nx//2]);
    ix1 = min(ix[ix>nx//2]);
    iy = numpy.where(psf[nx//2,:]<0)[0];
    iy0 = max(iy[iy<ny//2]);
    iy1 = min(iy[iy>ny//2]);
    print ix0,ix1,iy0,iy1;
    psf = psf[ix0:ix1,iy0:iy1];
  psf[psf<0] = 0;

  # estimate gaussian parameters, then fit
  import gaussfitter2
  parms0 = gaussfitter2.moments(psf,circle=0,rotate=1,vheight=0);
  print parms0;
  dprint(2,"Estimated parameters are",parms0);
  parms = gaussfitter2.gaussfit(psf,None,parms0,autoderiv=1,return_all=0,circle=0,rotate=1,vheight=0);
  dprint(0,"Fitted parameters are",parms);

  # now swap x and y around, since our axes are in reverse order
  ampl,y0,x0,sy,sx,rot = parms;

  # get pixel sizes in radians (by constructing a projection object)
  proj = Projection.FITSWCS(hdr);
  xscale,yscale = proj.xscale,proj.yscale;

  sx_rad = abs(sx * proj.xscale);
  sy_rad = abs(sy * proj.yscale);
  rot -= 90;  # convert West through North PA into the conventional North through East
  if sx_rad < sy_rad:
    sx_rad,sy_rad = sy_rad,sx_rad;
    rot -= 90;
  rot %= 180;

  dprintf(1,"Fitted gaussian PSF FWHM of %f x %f pixels (%f x %f arcsec), PA %f deg\n",sx*FWHM,sy*FWHM,sx_rad*FWHM*ARCSEC,sy_rad*FWHM*ARCSEC,rot);

  return sx_rad,sy_rad,rot/DEG;

def convolveGaussian (x1,y1,p1,x2,y2,p2):
    """convolves a Gaussian with extents x1,y1 and position angle p1
    with another Gaussian given by x2,y2,p2, and returns the extents 
    and angle of the resulting Gaussian."""
    # convert to Fourier plane extents, FT transforms a -> pi^2/a
    u1,v1,u2,v2 = [ (math.pi**2)*2*a**2 for a in x1,y1,x2,y2 ];
#    print "uv coeffs",u1,v1,u2,v2;
    c1,s1 = math.cos(p1),math.sin(p1);
    c2,s2 = math.cos(p2),math.sin(p2);
    # in the FT, this is a product of two Gaussians, each of the form:
    #   exp(-( u*(cx+sy)^2 + v*(cx-sy)^2))
    # note how we rotate BACK through the position angle
    # The product is necessarily a Gaussian itself, of the form
    #   exp(-(a.u^2+2b.u.v+c.v^2))
    # So we just need to collect the rotated Gaussian coefficients into a, b and c
    a = u1*c1**2+v1*s1**2+u2*c2**2+v2*s2**2
    c = u1*s1**2+v1*c1**2+u2*s2**2+v2*c2**2
    b = c1*s1*(u1-v1)+c2*s2*(u2-v2)
#    print "a,b,c",a,b,c;
    # ok, find semi-major axes a1, b1 using the formula from http://mathworld.wolfram.com/Ellipse.html eq. 21-22
    # to go from a general quadratic curve (with a,b,c given above, d=f=0, g=-1) to semi-axes a',b'
    D = math.sqrt((a-c)**2+4*b**2)
    E = a+c
    a1 = math.sqrt(2/(E-D))
    b1 = math.sqrt(2/(E+D))
#    print "a',b'",a1,b1,"coeffs",1/(a1**2),1/(b1**2)
    # and derive rotation angle
    if b:
        p1 = math.atan2(2*b,a-c)/2 + math.pi/2
#        if a > c:
#          p1 += math.pi/2
    else:
        p1 = 0 if a <= c else math.pi/2
#    print "rotation",p1/DEG
    # ok, convert a1,b1 from uv-plane to image plane 
    x1 = math.sqrt(1/(2*math.pi**2*a1**2))
    y1 = math.sqrt(1/(2*math.pi**2*b1**2))
    # note that because of reciprocality, y1 becomes the major axis and x1 the minor axis, so adjust for that
    return y1,x1,(p1-math.pi/2)%math.pi;  
  
def getImageCube (fitshdu,filename="",extra_axes=None):
  """Converts a FITS HDU (consisting of a header and data) into a 4+-dim numpy array where the
  first two axes are x and y, the third is Stokes (possibly of length 1, if missing in the
  original image), and the rest are either as found in the FITS header (if extra_axes=None),
  or in the order specified by CTYPE in extra_axes (if present, else a dummy axis of size 1 is inserted),
  with axes not present in extra_axes removed by taking the 0-th plane along each.
  Returns tuple of
    array,stokes_list,extra_axes_ctype_list,removed_axes_ctype_list
  e.g. array,("I","Q"),("FREQ--FOO","TIME--BAR")
  """
  hdr = fitshdu.header;
  data = fitshdu.data;
  # recognized axes
  ix = iy = istokes = None;
  naxis = len(data.shape);
  # other axes which will be returned
  other_axes = [];
  other_axes_ctype = [];
  remove_axes = [];
  remove_axes_ctype = [];
  # match axis ctype
  # this makes FREQ equivalent to FELO*
  def match_ctype (ctype,ctype_list):
    for i,ct in enumerate(ctype_list):
      if ct == ctype or ( ct == "FREQ" and ctype.startswith("FELO") ) or ( ctype == "FREQ" and ct.startswith("FELO") ):
        return i;
    return None;
  # identify X, Y and stokes axes
  for n in range(naxis):
    iax = naxis-1-n;
    axs = str(n+1);
    ctype = hdr.get('CTYPE'+axs).strip().upper();
    if ix is None and FITSHeaders.isAxisTypeX(ctype):
      ix = iax; # in numpy order, axes are reversed
    elif iy is None and FITSHeaders.isAxisTypeY(ctype):
      iy = iax;
    elif ctype == 'STOKES':
      if istokes is not None:
        raise ValueError,"duplicate STOKES axis in FITS file %s"%filename;
      istokes = iax;
      crval = hdr.get('CRVAL'+axs,0);
      cdelt = hdr.get('CDELT'+axs,1);
      crpix = hdr.get('CRPIX'+axs,1)-1;
      values = map(int,list(crval + (numpy.arange(data.shape[iax]) - crpix)*cdelt));
      stokes_names = [ (FITSHeaders.StokesNames[i]
                        if i>0 and i<len(FITSHeaders.StokesNames) else "%d"%i) for i in values ];
    else:
      other_axes.append(iax);
      other_axes_ctype.append(ctype);
  # not found?
  if ix is None or iy is None:
    raise ValueError,"FITS file %s does not appear to contain an X and/or Y axis"%filename;
  # form up shape of resulting image, and order of axes for transpose
  shape = [data.shape[ix],data.shape[iy]];
  axes = [ix,iy];
  # add stokes axis
  if istokes is None:
    shape.append(1);
    stokes_names = ("I",);
  else:
    shape.append(data.shape[istokes]);
    axes.append(istokes);
  if extra_axes:
    # if a fixed order for the extra axes is specified, add the ones we found
    for ctype in extra_axes:
      i = match_ctype(ctype,other_axes_ctype);
      if i is not None:
        iax = other_axes[i];
        axes.append(iax);
        shape.append(data.shape[iax]);
      else:
        shape.append(1);
    # add the ones that were not found into the remove list
    for iaxis,ctype in zip(other_axes,other_axes_ctype):
      if match_ctype(ctype,extra_axes) is None:
        axes.append(iaxis);
        remove_axes.append(iaxis);
        remove_axes_ctype.append(ctype);
  # return all extra axes found in header
  else:
    shape += [ data.shape[i] for i in other_axes ];
    axes += other_axes;
    extra_axes = other_axes_ctype;
  # tranpose
  data = data.transpose(axes);
  # trim off axes which are to be removed, if we have any
  if remove_axes:
    data = data[[Ellipsis]+[0]*len(remove_axes)];
  # reshape and return
  return data.reshape(shape),stokes_names,extra_axes,remove_axes_ctype;


class ImageResampler (object):
  """This class resamples images from one projection ("source") to another ("target").""";
  def __init__(self,sproj,tproj,sl,sm,tl,tm):
    """Creates resampler.
    sproj,tproj are the source and target Projection objects.
    sl,sm is a (sorted, ascending) list of l,m coordinates in the source image
    tl,tm is a (sorted, ascending) list of l,m coordinates in the target image
    """
    # convert tl,tm to to source coordinates
    # find the overlap region first, to keeps the number of coordinate conversions to a minimum
    overlap = astLib.astWCS.findWCSOverlap(sproj.wcs,tproj.wcs);
    tx2,tx1,ty1,ty2 = overlap['wcs2Pix'];
    # no overlap? stop then
    if tx1 > tl[-1] or tx2 < tl[0] or ty1 > tm[-1] or ty2 < tm[0]:
      self._target_slice = None,None;
      return;
    tx1 = max(0,int(math.floor(tx1)));
    tx2 = min(len(tl),int(math.floor(tx2+1)));
    ty1 = max(0,int(math.floor(ty1)));
    ty2 = min(len(tm),int(math.floor(ty2+1)));
    tl = tl[tx1:tx2];
    tm = tm[ty1:ty2];
    dprint(4,"overlap target pixels are %d:%d and %d:%d"%(tx1,tx2,ty1,ty2));

    #### The code below works but can be very slow  (~minutes) when doing large images, because of WCS
    ## make target lm matrix
    #tmat = numpy.zeros((2,len(tl),len(tm)));
    #tmat[0,...] = tl[:,numpy.newaxis];
    #tmat[1,...] = tm[numpy.newaxis,:];
    ## convert this to radec. Go through list since that's what Projection expects
    #dprint(4,"converting %d target l/m pixel coordinates to radec"%(len(tl)*len(tm)));
    #ra,dec = tproj.radec(tmat[0,...].ravel(),tmat[1,...].ravel())
    #dprint(4,"converting radec to source l/m");
    #tls,tms = sproj.lm(ra,dec);
    #tmat[0,...] = tls.reshape((len(tl),len(tm)));
    #tmat[1,...] = tms.reshape((len(tl),len(tm)));

    #### my alternative conversion code
    ## source to target is always an affine transform (one image projected into the plane of another, right?), so
    ## use WCS to map the corners, and figure out a linear transform from there

    # this maps three corners
    t00 = sproj.lm(*tproj.radec(tl[0],tm[0]));
    t1x = sproj.lm(*tproj.radec(tl[-1],tm[0]));
    t1y = sproj.lm(*tproj.radec(tl[0],tm[-1]));

    tmat = numpy.zeros((2,len(tl),len(tm)));
    tlnorm = (tl-tl[0])/(tl[-1]-tl[0]);
    tmnorm = (tm-tm[0])/(tm[-1]-tm[0]);
    tmat[0,...] = t00[0] + (tlnorm*(t1x[0]-t00[0]))[:,numpy.newaxis] + (tmnorm*(t1y[0]-t00[0]))[numpy.newaxis,:];
    tmat[1,...] = t00[1] + (tmnorm*(t1y[1]-t00[1]))[numpy.newaxis,:] + (tlnorm*(t1x[1]-t00[1]))[:,numpy.newaxis];

    dprint(4,"setting up slices");
    # ok, now find pixels in tmat that are within the source image extent
    tmask = (sl[0]<=tmat[0,...])&(tmat[0,...]<=sl[-1])&(sm[0]<=tmat[1,...])&(tmat[1,...]<=sm[-1]);
    # find extents along target's l and m axis
    # tmask_l/m is true for each target column/row that has pixels within the source image
    tmask_l = numpy.where(tmask.sum(1)>0)[0];
    tmask_m = numpy.where(tmask.sum(0)>0)[0];
    # check if there's no overlap at all -- return then
    if not len(tmask_l) or not len(tmask_m):
      self._target_slice = None,None;
      return;
    # ok, now we know over which pixels of the target image need to be interpolated
    ix0,ix1 = tmask_l[0],tmask_l[-1]+1;
    iy0,iy1 = tmask_m[0],tmask_m[-1]+1;
    self._target_slice = slice(ix0+tx1,ix1+tx1),slice(iy0+ty1,iy1+ty1);
    dprint(4,"slices are",ix0,ix1,iy0,iy1);
    # make [2,nx,ny] array of interpolation coordinates
    self._target_coords = tmat[:,ix0:ix1,iy0:iy1];

  def targetSlice (self):
    return self._target_slice;

  def __call__ (self,image):
    if self._target_slice[0] is None:
      return 0;
    else:
      return map_coordinates(image,self._target_coords);

def restoreSources (fits_hdu,sources,gmaj,gmin=None,grot=0,freq=None,primary_beam=None,apply_beamgain=False,ignore_nobeam=False):
  """Restores sources (into the given FITSHDU) using a Gaussian PSF given by gmaj/gmin/grot, in radians.
  gmaj/gmin is major/minor sigma parameter; grot is PA in the North thru East convention (PA=0 is N).

  If gmaj=0, uses delta functions instead.
  If freq is specified, converts flux to the specified frequency.
  If primary_beam is specified, uses it to apply a PB gain to each source. This must be a function of two arguments:
  r and freq, returning the power beam gain.
  If apply_beamgain is true, applies beamgain atribute instead, if this exists.
  Source tagged 'nobeam' will not have the PB gain applied, unless ignore_nobeam=True
  """;
  hdr = fits_hdu.header;
  data,stokes,extra_data_axes,dum = getImageCube(fits_hdu);
  # create projection object, using pixel coordinates
  proj = Projection.FITSWCSpix(hdr);
  naxis = len(data.shape);
  nx = data.shape[0];
  ny = data.shape[1];
  dprintf(1,"Read image of shape %s\n",data.shape);
  # Now we make "indexer" tuples. These use the numpy.newarray index to turn elementary vectors into
  # full arrays of the same number of dimensions as 'data' (data can be 2-, 3- or 4-dimensional, so we need
  # a general solution.)
  # For e.g. a nfreq x nstokes x ny x nx array, the following objects are created:
  #   x_indexer    turns n-vector vx into a _,_,_,n array
  #   y_indexer    turns m-vector vy into a _,_,m,_ array
  #   stokes_indexer turns the stokes vector into a _,nst,_,_ array
  # ...where "_" is numpy.newaxis.
  # The happy result of all this is that we can add a Gaussian into the data array at i1:i2,j1:j2 as follows:
  #  1. form up vectors of world coordinates (vx,vy) corresponding to pixel coordinates i1:i2 and j1:j2
  #  2. form up vector of Stokes parameters
  #  3. g = Gauss(vx[x_indexer],vy[y_indexer])*stokes[stokes_indexer]
  #  4. Just say data[j1:j2,i1:2,...] += g
  # This automatically expands all array dimensions as needed.

  # This is a helper function, returns an naxis-sized tuple, with slice(None) in the Nth
  # position, and elem_index elsewhere.
  def make_axis_indexer (n,elem_index=numpy.newaxis):
    indexer = [elem_index]*naxis;
    indexer[n] = slice(None);
    return tuple(indexer);
  x_indexer = make_axis_indexer(0);
  y_indexer = make_axis_indexer(1);
  # figure out stokes
  nstokes = len(stokes);
  stokes_vec = numpy.zeros((nstokes,));
  stokes_indexer = make_axis_indexer(2);
  dprint(2,"Stokes are",stokes);
  dprint(2,"Stokes indexing vector is",stokes_indexer);
  # get pixel sizes, in radians
  # gmaj != 0: use gaussian. Estimate PSF box size. We want a +/-5 sigma box
  if gmaj > 0:
    # convert grot from N-E to W-N (which is the more conventional mathematical definition of these things), so X is major axis
    grot += math.pi/2;  
    if gmin == 0:
      gmin = gmaj;
    cos_rot = math.cos(grot);
    sin_rot = math.sin(-grot);  # rotation is N->E, so swap the sign
  else:
    gmaj = gmin = grot = 0;
  conv_kernels = {};
  # loop over sources in model
  for src in sources:
    # get normalized intensity, if spectral info is available
    if freq is not None and getattr(src,'spectrum',None):
      ni = src.spectrum.normalized_intensity(freq);
      dprintf(3,"Source %s: normalized spectral intensity is %f\n",src.name,ni);
    else:
      ni = 1;
    #  multiply that by PB gain, if given
    if ignore_nobeam or not getattr(src,'nobeam',False):
      if apply_beamgain and hasattr(src,'beamgain'):
        ni *= getattr(src,'beamgain');
      elif primary_beam:
        r = getattr(src,'r',None);
        if r is not None:
          pb = primary_beam(r,freq);
          ni *= pb;
        dprintf(3,"Source %s: r=%g pb=%f, normalized intensity is %f\n",src.name,r,pb,ni);
    # process point sources
    if src.typecode in ('pnt','Gau'):
      # pixel coordinates of source
      xsrc,ysrc = proj.lm(src.pos.ra,src.pos.dec);
      # form up stokes vector
      for i,st in enumerate(stokes):
         stokes_vec[i] = getattr(src.flux,st,0)*ni;
      dprintf(3,"Source %s, %s Jy, at pixel %f,%f\n",src.name,stokes_vec,xsrc,ysrc);
      # for gaussian sources, convolve with beam
      if src.typecode == 'Gau':
        pa0 = src.shape.pa+math.pi/2;  # convert PA from N->E to conventional W->N
        ex0,ey0 = src.shape.ex/FWHM,src.shape.ey/FWHM;  # convert extents from FWHM to sigmas, since gmaj/gmin is in same scale
        if gmaj > 0:
          ex,ey,pa = convolveGaussian(ex0,ey0,pa0,gmaj,gmin,grot);
          # normalize flux by beam/extent ratio
          stokes_vec *= (gmaj*gmin)/(ex*ey);
          #print "%3dx%-3d@%3d * %3dx%-3d@%3d -> %3dx%-3d@%3d"%(
            #ex0 *FWHM*ARCSEC,ey0 *FWHM*ARCSEC,(pa0-math.pi/2)*DEG,
            #gmaj*FWHM*ARCSEC,gmin*FWHM*ARCSEC,(grot-math.pi/2)*DEG,
            #ex  *FWHM*ARCSEC,ey  *FWHM*ARCSEC,(pa-math.pi/2)*DEG);
        else:
          # normalize flux by pixel/extent ratio
          ex,ey,pa = ex0,ey0,pa0;
          stokes_vec *= (abs(proj.xscale*proj.yscale))/(ex*ey);
      else:
        ex,ey,pa = gmaj,gmin,grot;
      # gmaj != 0: use gaussian.
      if ex > 0 or ey > 0:
        # work out restoring box
        box_radius = 5*(max(ex,ey))/min(abs(proj.xscale),abs(proj.yscale));
        dprintf(2,"Will use a box of radius %f pixels for restoration\n",box_radius);
        cos_pa = math.cos(pa);
        sin_pa = math.sin(-pa);  # rotation is N->E, so swap the sign
        # pixel coordinates of box around source in which we evaluate the gaussian
        i1 = max(0,int(math.floor(xsrc-box_radius)));
        i2 = min(nx,int(math.ceil(xsrc+box_radius)));
        j1 = max(0,int(math.floor(ysrc-box_radius)));
        j2 = min(ny,int(math.ceil(ysrc+box_radius)));
        # skip sources if box doesn't overlap image
        if i1>=i2 or j1>=j2:
          continue;
        # now we convert pixel indices within the box into world coordinates, relative to source position
        xi = (numpy.arange(i1,i2) - xsrc)*proj.xscale;
        yj = (numpy.arange(j1,j2) - ysrc)*proj.yscale;
        # work out rotated coordinates
        xi1 = (xi*cos_pa)[x_indexer] - (yj*sin_pa)[y_indexer];
        yj1 = (xi*sin_pa)[x_indexer] + (yj*cos_pa)[y_indexer];
        # evaluate gaussian at these, scale up by stokes vector
        gg = stokes_vec[stokes_indexer]*numpy.exp(-((xi1/ex)**2+(yj1/ey)**2)/2.);
        # add into data
        data[i1:i2,j1:j2,...] += gg;
      # else gmaj=0: use delta functions
      else:
        xsrc = int(round(xsrc));
        ysrc = int(round(ysrc));
        # skip sources outside image
        if xsrc < 0 or xsrc >= nx or ysrc < 0 or ysrc >= ny:
          continue;
        xdum = numpy.array([1]);
        ydum = numpy.array([1]);
        data[xsrc:xsrc+1,ysrc:ysrc+1,...] += stokes_vec[stokes_indexer]*xdum[x_indexer]*ydum[y_indexer];
    # process model images -- convolve with PSF and add to data
    elif src.typecode == "FITS":
      modelff = pyfits.open(src.shape.filename);
      model,model_stokes,extra_model_axes,removed_model_axes = \
          getImageCube(modelff[0],src.shape.filename,extra_axes=extra_data_axes);
      modelproj = Projection.FITSWCSpix(modelff[0].header);
      # map Stokes planes: at least the first one ("I", presumably) must be present
      # The rest are represented by indices in model_stp. Thus e.g. for an IQUV data image and an IV model,
      # model_stp will be [0,-1,-1,1]
      model_stp = [ (model_stokes.index(st) if st in model_stokes else -1) for st in stokes ];
      if model_stp[0] < 0:
        print "Warning: model image %s lacks Stokes %s, skipping."%(src.shape.filename,model_stokes[0]);
        continue;
      # figure out whether the images overlap at all
      # in the trivial case, both images have the same WCS, so no resampling is needed
      if model.shape[:2] == data.shape[:2] and modelproj == proj:
        model_resampler = lambda x:x;
        data_x_slice = data_y_slice = slice(None);
        dprintf(3,"Source %s: same resolution as output, no interpolation needed\n",src.shape.filename);
      # else make a resampler engine
      else:
        model_resampler = ImageResampler(modelproj,proj,
          numpy.arange(model.shape[0],dtype=float),numpy.arange(model.shape[1],dtype=float),
          numpy.arange(data.shape[0],dtype=float),numpy.arange(data.shape[1],dtype=float));
        data_x_slice,data_y_slice = model_resampler.targetSlice();
        dprintf(3,"Source %s: resampling into image at %s, %s\n",src.shape.filename,data_x_slice,data_y_slice);
        # skip this source if no overlap
        if data_x_slice is None or data_y_slice is None:
          continue;
      # warn about ignored model axes (e.g. when model has frequency and our output doesn't)
      if removed_model_axes:
        print "Warning: model image %s has one or more axes that are not present in the output image:"%src.shape.filename;
        print "  taking the first plane along (%s)."%(",".join(removed_model_axes));
      # evaluate convolution kernel for this model scale, if not already cached
      conv_kernel = conv_kernels.get((modelproj.xscale,modelproj.yscale),None);
      if conv_kernel is None:
        box_radius = 5*(max(gmaj,gmin))/min(abs(modelproj.xscale),abs(modelproj.yscale));
        radius = int(round(box_radius));
        # convert pixel coordinates into world coordinates relative to 0
        xi = numpy.arange(-radius,radius+1)*modelproj.xscale
        yj = numpy.arange(-radius,radius+1)*modelproj.yscale
        # work out rotated coordinates
        xi1 = (xi*cos_rot)[:,numpy.newaxis] - (yj*sin_rot)[numpy.newaxis,:];
        yj1 = (xi*sin_rot)[:,numpy.newaxis] + (yj*cos_rot)[numpy.newaxis,:];
        # evaluate convolution kernel
        conv_kernel = numpy.exp(-((xi1/gmaj)**2+(yj1/gmin)**2)/2.);
        conv_kernels[modelproj.xscale,modelproj.yscale] = conv_kernel;
      # Work out data slices that we need to loop over.
      # For every 2D slice in the data image cube (assuming other axes besides x/y), we need to apply a
      # convolution to the corresponding model slice, and add it in to the data slice. The complication
      # is that any extra axis may be of length 1 in the model and of length N in the data (e.g. frequency axis),
      # in which case we need to add the same model slice to all N data slices. The loop below puts together a series
      # of index tuples representing each per-slice operation.
      # These two initial slices correspond to the x/y axes. Additional indices will be appended to these in a loop
      slices0 = [([data_x_slice,data_y_slice],[slice(None),slice(None)])];
      # work out Stokes axis
      sd0 = [data_x_slice,data_y_slice];
      sm0 = [slice(None),slice(None)];
      slices = [];
      slices = [ (sd0+[dst],sm0+[mst]) for dst,mst in enumerate(model_stp) if mst >= 0 ];
      #for dst,mst in enumerate(model_stp):
        #if mst >= 0:
          #slices = [ (sd0+[dst],sm0+[mst]) for sd0,sm0 in slices ];
      # now loop over extra axes
      for axis in range(3,len(extra_data_axes)+3):
        # list of data image indices to iterate over for this axis, 0...N-1
        indices = [[x] for x in range(data.shape[axis])];
        # list of model image indices to iterate over
        if model.shape[axis] == 1:
          model_indices = [[0]]*len(indices);
        # shape-n: must be same as data, in which case 0..N-1 is assigned to 0..N-1
        elif model.shape[axis] == data.shape[axis]:
          model_indices = indices;
        # else error
        else:
          raise RuntimeError,"axis %s of model image %s doesn't match that of output image"%\
                              (extra_data_axes[axis-3],src.shape.filename);
        # update list of slices
        slices =[ (sd0+sd,si0+si) for sd0,si0 in slices for sd,si in zip(indices,model_indices) ];
      # now loop over slices and assign
      for sd,si in slices:
        conv = convolve(model[tuple(si)],conv_kernel);
        data[tuple(sd)] += model_resampler(conv);
        ## for debugging these are handy:
        #data[0:conv.shape[0],0:conv.shape[1],0,0] = conv;
        #data[0:conv_kernel.shape[0],-conv_kernel.shape[1]:,0,0] = conv_kernel;



# -*- coding: utf-8 -*-
#
#% $Id: BBS.py 8378 2011-08-30 15:18:30Z oms $ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys
import traceback
import math
import struct
import time
import os.path
import re

import  numpy
import pyfits

import Kittens.utils

import Tigger.Models.Formats
from Tigger.Models import ModelClasses
from Tigger.Models import SkyModel
from Tigger import Coordinates
from Tigger.Models.Formats import dprint,dprintf
from math import cos,sin,acos,asin,atan2,sqrt,pi

DEG = math.pi/180
ARCMIN = DEG/60
ARCSEC = ARCMIN/60

"""
Loads an AIPS-format clean component list
"""

def lm_to_radec (l,m,ra0,dec0):
  """Returns ra,dec corresponding to l,m w.r.t. direction ra0,dec0""";
  # see formula at http://en.wikipedia.org/wiki/Orthographic_projection_(cartography)
  rho = sqrt(l**2+m**2);
  if rho == 0.0:
    ra = ra0
    dec = dec0
  else:
    cc = asin(rho);
    ra = ra0 + atan2( l*sin(cc),rho*cos(dec0)*cos(cc)-m*sin(dec0)*sin(cc) );
    dec = asin( cos(cc)*sin(dec0) + m*sin(cc)*cos(dec0)/rho );
  return ra,dec;

_units = dict(DEG=DEG, DEGREE=DEG, DEGREES=DEG,
              RAD=1, RADIAN=1, RADIANS=1,
              ARCMIN=ARCMIN, ARCMINS=ARCMIN,
              ARCSEC=ARCSEC, ARCSECS=ARCSEC
              )

def load (filename,center=None,**kw):
  """Imports an AIPS clean component list from FITS table
  """
  srclist = [];
  dprint(1,"importing AIPS clean component FITS table",filename);
  # read file
  ff = pyfits.open(filename);
  
  if center is None:
    hdr = ff[0].header
    ra  = hdr['CRVAL1'] * _units[hdr.get('CUNIT1','DEG').strip()]
    dec = hdr['CRVAL2'] * _units[hdr.get('CUNIT2','DEG').strip()]

    print "Using FITS image centre (%.4f, %.4f deg) as field centre" % (ra/DEG, dec/DEG)
    center = ra, dec

  # now process file line-by-line
  cclist = ff[1].data;
  hdr = ff[1].header
  ux = _units[hdr.get('TUNIT2','DEG').strip()]
  uy = _units[hdr.get('TUNIT3','DEG').strip()]
  for num,ccrec in enumerate(cclist):
    stokes_i,dx,dy = map(float,ccrec);
    # convert dx/dy to real positions
    l,m = sin(dx*ux), sin(dy*uy);
    ra,dec = lm_to_radec(l,m,*center);
    pos = ModelClasses.Position(ra,dec);
    flux = ModelClasses.Flux(stokes_i);
    # now create a source object
    src = SkyModel.Source('cc%d'%num,pos,flux);
    src.setAttribute('r',math.sqrt(l*l+m*m));
    srclist.append(src);
  dprintf(2,"imported %d sources from file %s\n",len(srclist),filename);
  # create model
  model = ModelClasses.SkyModel(*srclist);
  # setup model center
  model.setFieldCenter(*center);
  # setup radial distances
  projection = Coordinates.Projection.SinWCS(*model.fieldCenter());
  return model;


Tigger.Models.Formats.registerFormat("AIPSCCFITS",load,"AIPS CC FITS model",(".fits",".FITS",".fts",".FTS"));



# -*- coding: utf-8 -*-
#
#% $Id$
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import sys,traceback,math,numpy,re

import Kittens.utils

from Tigger.Models import ModelClasses
from Tigger.Models import SkyModel
from Tigger import Coordinates
import Tigger.Models.Formats
from Tigger.Models.Formats import dprint,dprintf


DefaultDMSFormat = dict(name=0,
    ra_h=1,ra_m=2,ra_s=3,dec_d=4,dec_m=5,dec_s=6,
    i=7,q=8,u=9,v=10,spi=11,rm=12,emaj_s=13,emin_s=14,pa_d=15,
    freq0=16,tags=slice(17,None));

DefaultDMSFormatString = "name ra_h ra_m ra_s dec_d dec_m dec_s i q u v spi rm emaj_s emin_s pa_d freq0 tags...";

FormatHelp = """
ASCII files are treated as columns of whitespace-separated values. The order
of the columns is determined by a format string, which can be specified in
the first line of the file (prefixed by "#format:"), or supplied by the
user.  Note that in subsequent lines the "#" character is treated as a
comment delimiter, everything following a "#" is ignored.

The format string contains a simple list of field names, such as "name ra_d
dec_d i".  Fields with unrecognized names are simply ignored -- a good way
to skip over unwanted columns is to use a name like 'dummy' or '-'.

The following field names are recognized. Note that only a subset of these
needs to be present (as a minimum, coordinates and I flux needs to be
supplied, but the rest is optional):

name:             source name
ra_{rad,d,h,m,s}: RA or RA component,
                  (in radians, degrees, hours, minutes or seconds)
ra_err_{rad,d,h,m,s}: error on RA (in appropriate units)
dec_{rad,d,m,s}:  declination or declination component
dec_sign:         declination sign (+ or -)
dec_err_{rad,d,m,s}: error on dec (in appropriate units)
i,q,u,v:          IQUV fluxes
{i,q,u,v}_err:    errors on fluxes
pol_frac:         linear polarization fraction
                  (will interpret both "0.1" and "10%" correctly)
pol_pa_{rad,d}:   linear polarization angle
rm:               rotation measure (freq0 must be supplied as well)
rm_err:           error on rotation measures
spi:              spectral index (freq0 must be supplied as well)
spi2,3,4...:      spectral curvature
spi_err,spi2_err,...: error on spectral index and curvature
freq0:            reference frequency, for rm and/or spi
emaj_{rad,d,m,s}: source extent, major axis (for Gaussian sources)
emin_{rad,d,m,s}: source extent, minor axis (for Gaussian sources)
{emin,emaj}_err_{rad,d,m,s}:  error on source extent
pa_{rad,d}:       position angle (for Gaussian sources)
pa_err_{ra,d}:    error on position angle
tags:             comma-separated source tags
tags...:          absorb all remaining fields as source tags
:TYPE:ATTR        custom attribute. Contents of field will be converted to Python TYPE
                  (bool, int, float, complex, str) and associated with custom source atribute "ATTR"
""";

DEG = math.pi/180;

# dict of angulr units with their scale in radians
ANGULAR_UNITS = dict(rad=1,d=DEG,m=DEG/60,s=DEG/3600,h=DEG*15)
# subsets of angular units for leading RA or Dec column
ANGULAR_UNITS_RA = dict(rad=1,d=DEG,h=DEG*15)
ANGULAR_UNITS_DEC = dict(rad=1,d=DEG)

def load (filename,format=None,freq0=None,center_on_brightest=False,min_extent=0,verbose=0,**kw):
  """Imports an ASCII table
  The 'format' argument can be either a dict (such as the DefaultDMSFormat dict above), or a string such as DefaultDMSFormatString.
  (Other possible field names are "ra_d", "ra_rad", "dec_rad", "dec_sign".)
  If None is specified, DefaultDMSFormat is used.
  The 'freq0' argument supplies a default reference frequency (if one is not contained in the file.)
  If 'center_on_brightest' is True, the mpodel field center will be set to the brightest source.
  'min_extent' is minimal source extent (in radians), above which a source will be treated as a Gaussian rather than a point component.
  """
  srclist = [];
  dprint(1,"importing ASCII DMS file",filename);
  # brightest source and its coordinates
  maxbright = 0;
  brightest_name = radec0 = None;

  # Get column number associated with field from format dict, as well as the error
  # column number. Returns tuple of indices, with None index indicating no such column
  def get_field (name):
    return format.get(name,None),format.get(name+"_err",None);
  # Get column number associated with field from format dict, as well as the error
  # column number. Field is an angle thus will be suffixed with _{rad,d,h,m,s}.
  # Returns tuple of
  #     column,scale,err_column,err_scale
  # with None index indicating no such column. Scale is scaling factor to convert
  # quantity in column to radians
  def get_ang_field (name,units=ANGULAR_UNITS):
    column = err_column = colunit = errunit = None
    units = units or ANGULAR_UNITS;
    for unit,scale in units.iteritems():
      if column is None:
        column = format.get("%s_%s"%(name,unit));
        if column is not None:
          colunit = scale;
      if err_column is None:
        err_column = format.get("%s_err_%s"%(name,unit))
        if err_column is not None:
          errunit = scale;
    return column,colunit,err_column,errunit;

  # helper function: returns element #num from the fields list, multiplied by scale, or None if no such field
  def getval (num,scale=1):
    return None if ( num is None or len(fields) <= num ) else float(fields[num])*scale;

  # now process file line-by-line
  linenum = 0;
  format_str = ''
  for line in file(filename):
    # for the first line, figure out the file format
    if not linenum:
      if not format and line.startswith("#format:"):
        format = line[len("#format:"):].strip();
        dprint(1,"file contains format header:",format);
      # set default format
      if format is None:
        format = DefaultDMSFormatString;
      # is the format a string rather than a dict? Turn it into a dict then
      if isinstance(format,str):
        format_str = format;
        # make list of fieldname,fieldnumber tuples
        fields = [ (field,i) for i,field in enumerate(format.split()) ];
        if not fields:
          raise ValueError,"illegal format string in file: '%s'"%format;
        # last fieldname can end with ... to indicate that it absorbs the rest of the line
        if fields[-1][0].endswith('...'):
          fields[-1] = (fields[-1][0][:-3],slice(fields[-1][1],None));
        # make format dict
        format = dict(fields);
      elif not isinstance(format,dict):
        raise TypeError,"invalid 'format' argument of type %s"%(type(format))
        # nf = max(format.itervalues())+1;
        # fields = ['---']*nf;
        # for field,number in format.iteritems():
        #   fields[number] = field;
        # format_str = " ".join(fields);
      # get list of custom attributes from format
      custom_attrs = [];
      for name,col in format.iteritems():
        if name.startswith(":"):
          m = re.match("^:(bool|int|float|complex|str):([\w]+)$",name);
          if not m:
            raise TypeError,"invalid field specification '%s' in format string"%name;
          custom_attrs.append((eval(m.group(1)),m.group(2),col));
      # get minimum necessary fields from format
      name_field = format.get('name',None);
      # flux
      i_field,i_err_field = get_field("i");
      if i_field is None:
        raise ValueError,"ASCII format specification lacks mandatory flux field ('i')";
      # main RA field
      ra_field,ra_scale,ra_err_field,ra_err_scale = get_ang_field('ra',ANGULAR_UNITS_RA);
      if ra_field is None:
        raise ValueError,"ASCII format specification lacks mandatory Right Ascension field ('ra_h', 'ra_d' or 'ra_rad')";
      # main Dec field
      dec_field,dec_scale,dec_err_field,dec_err_scale = get_ang_field('dec',ANGULAR_UNITS_DEC);
      if dec_field is None:
        raise ValueError,"ASCII format specification lacks mandatory Declination field ('dec_d' or 'dec_rad')";
      # polarization as QUV
      quv_fields = [ get_field(x) for x in ['q','u','v'] ];
      # linear polarization as fraction and angle
      polfrac_field = format.get('pol_frac',None);
      if polfrac_field is not None:
        polpa_field,polpa_scale = format.get('pol_pa_d',None),(math.pi/180);
        if not polpa_field is not None:
          polpa_field,polpa_scale = format.get('pol_pa_rad',None),1;
      # fields for extent parameters
      extent_fields = [ get_ang_field(x,ANGULAR_UNITS) for x in 'emaj','emin','pa' ];
      # all three must be present, else ignore
      if any( [ x[0] is None for x in extent_fields ] ):
        extent_fields = None;
      # fields for reference freq and RM and SpI
      freq0_field = format.get('freq0',None);
      rm_field,rm_err_field = get_field('rm');
      spi_fields = [ get_field('spi') ] + [ get_field('spi%d'%i) for i in range(2,10) ];
      tags_slice = format.get('tags',None);
    # now go on to process the line
    linenum += 1;
    try:
      # strip whitespace
      line = line.strip();
      dprintf(4,"%s:%d: read line '%s'\n",filename,linenum,line);
      # skip empty or commented lines
      if not line or line[0] == '#':
        continue;
      # split (at whitespace) into fields
      fields = line.split();
      # get  name
      name = fields[name_field] if name_field is not None else str(len(srclist)+1);
      i = getval(i_field);
      i_err = getval(i_err_field);
      # get position: RA
      ra = getval(ra_field);
      ra_err = getval(ra_err_field,ra_scale);
      if 'ra_m' in format:
        ra += float(fields[format['ra_m']])/60.;
      if 'ra_s' in format:
        ra += float(fields[format['ra_s']])/3600.;
      ra *= ra_scale;
      # position: Dec. Separate treatment of sign
      dec = abs(getval(dec_field));
      dec_err = getval(dec_err_field,dec_scale);
      if 'dec_m' in format:
        dec += float(fields[format['dec_m']])/60.;
      if 'dec_s' in format:
        dec += float(fields[format['dec_s']])/3600.;
      if fields[format.get('dec_sign',dec_field)][0] == '-':
        dec = -dec;
      dec *= dec_scale;
      # for up position object
      pos = ModelClasses.Position(ra,dec,ra_err=ra_err,dec_err=dec_err);
      # see if we have freq0

      # Use explicitly provided reference frequency for this source if available
      f0 = None
      if freq0_field is not None:
        try:
          f0 = float(fields[freq0_field])
          # If no default reference frequency for the model was supplied,
          # initialise from first source with a reference frequency
          if freq0 is None:
            freq0 = f0
            dprint(0,"Set default freq0 to %s "
                     "from source on line %s." % (f0, linenum));

        except IndexError:
          f0 = None

      # Otherwise use default reference frequency (derived from args
      # or first reference frequency found in source)
      if f0 is None and freq0 is not None:
        f0 = freq0

      # see if we have Q/U/V
      (q,q_err),(u,u_err),(v,v_err) = [ (getval(x),getval(x_err)) for x,x_err in quv_fields ];
      if polfrac_field is not None:
        pf = fields[polfrac_field];
        pf = float(pf[:-1])/100 if pf.endswith("%") else float(pf);
        ppa = float(fields[polpa_field])*polpa_scale if polpa_field is not None else 0;
        q = i*pf*math.cos(2*ppa);
        u = i*pf*math.sin(2*ppa);
        v = 0;
      # see if we have RM as well. Create flux object (unpolarized, polarized, polarized w/RM)
      rm,rm_err = getval(rm_field),getval(rm_err_field);
      if q is None:
        flux = ModelClasses.Polarization(i,0,0,0,I_err=i_err);
      elif f0 is None or rm is None:
        flux = ModelClasses.Polarization(i,q,u,v,I_err=i_err,Q_err=q_err,U_err=u_err,V_err=v_err);
      else:
        flux = ModelClasses.PolarizationWithRM(i,q,u,v,rm,f0,I_err=i_err,Q_err=q_err,U_err=u_err,V_err=v_err,rm_err=rm_err);
      # see if we have a spectral index
      if f0 is None:
        spectrum = None;
      else:
        spi = [ getval(x) for x,xerr in spi_fields ];
        spi_err = [ getval(xerr) for x,xerr in spi_fields ];
        dprint(4,name,"spi is",spi,"err is",spi_err)
        # if any higher-order spectral terms are specified, include them here but trim off all trailing zeroes
        while spi and not spi[-1]:
          del spi[-1];
          del spi_err[-1]
        if not spi:
          spectrum = None;
        elif len(spi) == 1:
          spectrum = ModelClasses.SpectralIndex(spi[0],f0);
          if spi_err[0] is not None:
            spectrum.spi_err = spi_err[0];
        else:
          spectrum = ModelClasses.SpectralIndex(spi,f0);
          if any([ x is not None for x in spi_err ]):
            spectrum.spi_err = spi_err;
      # see if we have extent parameters
      ex = ey = pa = 0;
      if extent_fields:
        ex,ey,pa = [ ( getval(x[0],x[1]) or 0 ) for x in extent_fields ];
        extent_errors = [ getval(x[2],x[3]) for x in extent_fields ];
      # form up shape object
      if (ex or ey) and max(ex,ey) >= min_extent:
        shape = ModelClasses.Gaussian(ex,ey,pa);
        for ifield,field in enumerate(['ex','ey','pa']):
          if extent_errors[ifield] is not None:
            shape.setAttribute(field+"_err",extent_errors[ifield]);
      else:
        shape = None;
      # get tags
      tagdict = {};
      if tags_slice:
        try:
          tags = fields[tags_slice];
        except IndexError:
          pass;
        for tagstr1 in tags:
          for tagstr in tagstr1.split(","):
            if tagstr[0] == "+":
              tagname,value = tagstr[1:],True;
            elif tagstr[0] == "-":
              tagname,value = tagstr[1:],False;
            elif "=" in tagstr:
              tagname,value = tagstr.split("=",1);
              if value[0] in "'\"" and value[-1] in "'\"":
                value = value[1:-1];
              else:
                try:
                  value = float(value);
                except:
                  continue;
            else:
              tagname,value = tagstr,True;
            tagdict[tagname] = value;
      # OK, now form up the source object
      # now create a source object
      dprint(3,name,ra,dec,i,q,u,v);
      src = SkyModel.Source(name,pos,flux,shape=shape,spectrum=spectrum,**tagdict);
      # get custom attributes
      for type_,attr,column in custom_attrs:
        if column is not None and len(fields) > column:
          src.setAttribute(attr,type_(fields[column]));
      # add to source list
      srclist.append(src);
      # check if it's the brightest
      brightness = src.brightness();
      if brightness > maxbright:
        maxbright = brightness;
        brightest_name = src.name;
        radec0 = ra,dec;
    except:
      if verbose:
        traceback.print_exc();
      dprintf(0,"%s:%d: %s, skipping\n",filename,linenum,str(sys.exc_info()[1]));
  dprintf(2,"imported %d sources from file %s\n",len(srclist),filename);
  # create model
  model = ModelClasses.SkyModel(*srclist);
  if freq0 is not None:
    model.setRefFreq(freq0);
  # set model format
  model.setAttribute("ASCII_Format",format_str);
  # setup model center
  if center_on_brightest and radec0:
    dprintf(2,"brightest source is %s (%g Jy) at %f,%f\n",brightest_name,maxbright,*radec0);
    model.setFieldCenter(*radec0);
  # setup radial distances
  projection = Coordinates.Projection.SinWCS(*model.fieldCenter());
  for src in model.sources:
    l,m = projection.lm(src.pos.ra,src.pos.dec);
    src.setAttribute('r',math.sqrt(l*l+m*m));
  return model;


def save (model,filename,sources=None,format=None,**kw):
  """
  Exports model to a text file
  """;
  if sources is None:
    sources = model.sources;
  dprintf(2,"writing %d model sources to text file %s\n",len(sources),filename);
  # create catalog parser based on either specified format, or the model format, or the default format
  format_str = format or getattr(model,'ASCII_Format',DefaultDMSFormatString);
  dprint(2,"format string is",format_str);
  # convert this into format dict
  fields = [ [field,i] for i,field in enumerate(format_str.split()) ];
  if not fields:
    raise ValueError,"illegal format string '%s'"%format;
  # last fieldname can end with ... ("tags..."), so strip it
  if fields[-1][0].endswith('...'):
    fields[-1][0] = fields[-1][0][:-3];
  # make format dict
  format = dict(fields);
  nfields = len(fields);
  # get minimum necessary fields from format
  name_field = format.get('name',None);
  # main RA field
  ra_rad_field,ra_d_field,ra_h_field,ra_m_field,ra_s_field = \
    [ format.get(x,None) for x in 'ra_rad','ra_d','ra_h','ra_m','ra_s' ];
  dec_rad_field,dec_d_field,dec_m_field,dec_s_field = \
    [ format.get(x,None) for x in 'dec_rad','dec_d','dec_m','dec_s' ];
  if ra_h_field is not None:
    ra_scale = 15;
    ra_d_field = ra_h_field;
  else:
    ra_scale = 1;
  # fields for reference freq and RM and SpI
  freq0_field = format.get('freq0',None);
  rm_field = format.get('rm',None);
  spi_field = format.get('spi',None);
  tags_field = format.get('tags',None);
  # open file
  ff = open(filename,mode="wt");
  ff.write("#format: %s\n"%format_str);
  # write sources
  nsrc = 0;
  for src in sources:
    # only write points and gaussians
    if src.shape is not None and not isinstance(src.shape,ModelClasses.Gaussian):
      dprint(3,"skipping source '%s': non-supported type '%s'"%(src.name,src.shape.typecode));
      continue;
    # prepare field values
    fval = ['0']*nfields;
    # name
    if name_field is not None:
      fval[name_field] = src.name;
    # position: RA
    ra,dec = src.pos.ra,src.pos.dec;
    # RA in radians
    if ra_rad_field is not None:
      fval[ra_rad_field] = str(ra);
    ra /= ra_scale;
    # RA in h/m/s or d/m/s
    if ra_m_field is not None:
      ra,ram,ras = src.pos.ra_hms_static(ra,scale=180,prec=1e-4);
      fval[ra_m_field] = str(ram);
      if ra_s_field is not None:
        fval[ra_s_field] = str(ras);
      if ra_d_field is not None:
        fval[ra_d_field] = str(ra);
    elif ra_d_field is not None:
        fval[ra_d_field] = str(ra*180/math.pi);
    # position: Dec
    if dec_rad_field is not None:
      fval[dec_rad_field] = str(dec);
    if dec_m_field is not None:
      dsign,decd,decm,decs = src.pos.dec_sdms();
      fval[dec_m_field] = str(decm);
      if dec_s_field is not None:
        fval[dec_s_field] = str(decs);
      if dec_d_field is not None:
        fval[dec_d_field] = dsign+str(decd);
    elif dec_d_field is not None:
        fval[dec_d_field] = str(dec*180/math.pi);
    # fluxes
    for stokes in "IQUV":
      field = format.get(stokes.lower());
      if field is not None:
        fval[field] = str(getattr(src.flux,stokes,0));
    # fractional polarization
    if 'pol_frac' in format:
      i,q,u = [ getattr(src.flux,stokes,0) for stokes in "IQU" ];
      fval[format['pol_frac']] = str(math.sqrt(q*q+u*u)/i);
      pa = math.atan2(u,q)/2;
      for field,scale in ('pol_pa_rad',1.),('pol_pa_d',DEG):
        ifield = format.get(field);
        if ifield is not None:
          fval[ifield] = str(pa/scale);
    # shape
    if src.shape:
      for parm,sparm in ("emaj","ex"),("emin","ey"),("pa","pa"):
        for field,scale in (parm,1.),(parm+'_rad',DEG),(parm+'_d',DEG),(parm+'_m',DEG/60),(parm+'_s',DEG/3600):
          ifield = format.get(field.lower());
          if ifield is not None:
            fval[ifield] = str(getattr(src.shape,sparm,0)/scale);
    # RM, spi, freq0
    if freq0_field is not None:
      freq0 = (src.spectrum and getattr(src.spectrum,'freq0',None)) or getattr(src.flux,'freq0',0);
      fval[freq0_field] = str(freq0);
    if rm_field is not None:
      fval[rm_field] = str(getattr(src.flux,'rm',0));
    if spi_field is not None and hasattr(src,'spectrum'):
      fval[spi_field] = str(getattr(src.spectrum,'spi',0));
    # tags
    if tags_field is not None:
      outtags = [];
      for tag,value in src.getTags():
        if isinstance(value,str):
          outtags.append("%s=\"%s\""%(tag,value));
        elif isinstance(value,bool):
          if value:
            outtags.append("+"+tag);
          else:
            outtags.append("-"+tag);
        elif isinstance(value,(int,float)):
          outtags.append("%s=%f"%(tag,value));
      fval[tags_field] = ",".join(outtags);
    # write the line
    ff.write(" ".join(fval)+"\n");
    nsrc += 1;

  ff.close();
  dprintf(1,"wrote %d sources to file %s\n",nsrc,filename);


Tigger.Models.Formats.registerFormat("ASCII",load,"ASCII table",(".txt",".lsm"),export_func=save);



# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2011
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


import sys
import traceback
import math
import struct
import time
import os.path

import  numpy

import Kittens.utils

import Tigger.Models.Formats
from Tigger.Models import ModelClasses
from Tigger.Models import SkyModel
from Tigger import Coordinates
from Tigger.Models.Formats import dprint,dprintf

def lm_ncp_to_radec (ra0,dec0,l,m):
  """Converts coordinates in l,m (NCP) relative to ra0,dec0 into ra,dec.""";
  sind0=math.sin(dec0)
  cosd0=math.cos(dec0)
  dl=l
  dm=m
  d0=dm*dm*sind0*sind0+dl*dl-2*dm*cosd0*sind0
  sind=math.sqrt(abs(sind0*sind0-d0))
  cosd=math.sqrt(abs(cosd0*cosd0+d0))
  if sind0>0:
    sind=abs(sind)
  else:
    sind=-abs(sind)
  dec=math.atan2(sind,cosd)
  if l != 0:
    ra=math.atan2(-dl,(cosd0-dm*sind0))+ra0
  else:
    ra=math.atan2((1e-10),(cosd0-dm*sind0))+ra0
  return ra,dec

def radec_to_lm_ncp (ra0,dec0,ra,dec):
  """Converts coordinates in l,m (NCP) relative to ra0,dec0 into ra,dec.""";
  l=-math.sin(ra-ra0)*math.cos(dec)
  sind0=math.sin(dec0)
  if sind0 != 0:
    m=-(math.cos(ra-ra0)*math.cos(dec)-math.cos(dec0))/math.sin(dec0)
  else:
    m=0
  return (l,m)


def parseGFH (gfh):
  """Parses the GFH (general file header?) structure at the beginning of the file""";
  ## type
  ftype = gfh[0:4].tostring()
  ## length & version
  fhlen,fver = struct.unpack('ii',gfh[4:12])
  ### creation date
  crdate = gfh[12:23].tostring()
  ### creation time
  crtime = gfh[23:28].tostring()
  ### revision date
  rrdate = gfh[28:39].tostring()
  ### revision time
  rrtime = gfh[39:44].tostring()
  ### revision count
  rcount = struct.unpack('i',gfh[44:48])
  rcount = rcount[0]
  ### node name
  nname = gfh[48:128].tostring()
  ### types
  dattp = struct.unpack('B',gfh[128:129])[0];
  link1,link2 = struct.unpack('ii',gfh[152:160]);
  ### the remaining info is not needed
  dprint(1,"read header type=%s, length=%d, version=%d, created=%s@%s, updated=%s@%s x %d, node name=%s, dattp=%d, link=%d,%d"%
    (ftype,fhlen,fver,crdate,crtime,rrdate,rrtime,rcount,nname,dattp,link1,link2));
  return (ftype,fhlen,fver,crdate,crtime,rrdate,rrtime,rcount,nname);

def parseMDH (mdh):
  """Parses the MDH (model file header?) structure""";
  maxlin,modptr,nsources,mtype = struct.unpack('iiii',mdh[12:28]);
  mepoch = struct.unpack('f',mdh[28:32])[0];
  ra0,dec0,freq0 = struct.unpack('ddd',mdh[32:56]);
  ### Max. # of lines in model or disk version
  ### pointer to model ???
  ### no of sources in model
  ### model type(0: no ra,dec, 1=app, 2=epoch)
  ### Epoch (e.g. 1950) if TYP=2 (float) : 4 bytes
  ###  Model centre RA (circles) : double
  ra0 *= math.pi*2;
  dec0 *= math.pi*2;
  ### Model centre FRQ (MHz)
  freq0 *= 1e6
  ### the remaining is not needed
  dprint(1,"read model header maxlines=%d, pointer=%d, sources=%d, type=%d, epoch=%f RA=%f, DEC=%f (rad) Freq=%f Hz"%
    (maxlin,modptr,nsources,mtype,mepoch,ra0,dec0,freq0));
  return (maxlin,modptr,nsources,mtype,mepoch,ra0,dec0,freq0);
  
def load (filename,import_src=True,import_cc=True,min_extent=0,**kw):
  """Imports a NEWSTAR MDL file.
  min_extent is minimal source extent (in radians), above which a source will be treated as a Gaussian rather than a point component.
  import_src=False causes source components to be omitted
  import_cc=False causes clean components to be omitted
  """;
  srclist = [];
  dprint(1,"importing NEWSTAR file",filename);
  # build the LSM from a NewStar .MDL model file
  # if only_cleancomp=True, only clean components are used to build the LSM
  # if no_cleancomp=True, no clean components are used to build the LSM
  ff = open(filename,mode="rb");
  
  ### read GFH and MDH headers -- 512 bytes
  try:
    gfh = numpy.fromfile(ff,dtype=numpy.uint8,count=512);
    mdh = numpy.fromfile(ff,dtype=numpy.uint8,count=64);
    # parse headers
    ftype,fhlen,fver,crdate,crtime,rrdate,rrtime,rcount,nname = parseGFH(gfh);
    if ftype != ".MDL":
      raise TypeError;
    maxlin,modptr,nsources,mtype,mepoch,ra0,dec0,freq0 = parseMDH(mdh);
    
    beam_const = 65*1e-9*freq0;
    
    ## temp dict to hold unique nodenames
    unamedict={}
    ### Models -- 56 bytes
    for ii in xrange(0,nsources):
      mdl = numpy.fromfile(ff,dtype=numpy.uint8,count=56)
      
      ### source parameters
      sI,ll,mm,id,sQ,sU,sV,eX,eY,eP,SI,RM = struct.unpack('fffiffffffff',mdl[0:48])
      ### type bits
      bit1,bit2 = struct.unpack('BB',mdl[52:54]);

      # convert fluxes
      sI *= 0.005    # convert from WU to Jy (1WU=5mJy)
      sQ *= sI;
      sU *= sI;
      sV *= sI;

      # Interpret bitflags 1: bit 0= extended; bit 1= Q|U|V <>0 and no longer used according to Wim
      fl_ext = bit1&1;
      # Interpret bitflags 2: bit 0= clean component; bit 3= beamed
      fl_cc = bit2&1;
      fl_beamed = bit2&8;

      ### extended source params: in arcsec, so multiply by ???
      if fl_ext:
        ## the procedure is NMOEXT in nscan/nmoext.for
        if eP == 0 and eX == eY:
          r0 = 0
        else:
          r0 = .5*math.atan2(-eP,eY-eX)
        r1 = math.sqrt(eP*eP+(eX-eY)*(eX-eY))
        r2 = eX+eY
        eX = 2*math.sqrt(abs(0.5*(r2+r1)))
        eY = 2*math.sqrt(abs(0.5*(r2-r1)))
        eP = r0

      # NEWSTAR MDL lists might have same source twice if they are
      # clean components, so make a unique name for them
      bname='N'+str(id);
      if unamedict.has_key(bname):
        uniqname = bname+'_'+str(unamedict[bname])
        unamedict[bname] += 1
      else:
        uniqname = bname
        unamedict[bname] = 1
      # compose source information
      pos = ModelClasses.Position(*lm_ncp_to_radec(ra0,dec0,ll,mm));
      flux  = ModelClasses.PolarizationWithRM(sI,sQ,sU,sV,RM,freq0);
      spectrum = ModelClasses.SpectralIndex(SI,freq0);
      tags = {};
      # work out beam gain and apparent flux
      tags['_lm_ncp'] = (ll,mm);
      tags['_newstar_r']   = tags['r'] = r = math.sqrt(ll*ll+mm*mm);
      tags['newstar_beamgain'] = bg = max(math.cos(beam_const*r)**6,.01);
      tags['newstar_id'] = id;
      if fl_beamed:
        tags['Iapp'] = sI*bg;
        tags['newstar_beamed'] = True;
        tags['flux_intrinsic'] = True;
      else:
        tags['flux_apparent'] = True;
      # make some tags based on model flags
      if fl_cc:
        tags['newstar_cc'] = True;
      # make shape if extended
      if fl_ext and max(eX,eY) >= min_extent:
        shape = ModelClasses.Gaussian(eX,eY,eP);
      else:
        shape = None;
      # compute apparent flux
      src = SkyModel.Source(uniqname,pos,flux,shape=shape,spectrum=spectrum,**tags);
      srclist.append(src);
  except:
    traceback.print_exc();
    raise TypeError("%s does not appear to be a valid NEWSTAR MDL file"%filename);
  
  dprintf(2,"imported %d sources from file %s\n",len(srclist),filename);
  return ModelClasses.SkyModel(ra0=ra0,dec0=dec0,freq0=freq0,pbexp='max(cos(65*1e-9*fq*r)**6,.01)',*srclist);


def save (model,filename,freq0=None,sources=None,**kw):
  """Saves model to a NEWSTAR MDL file.
  The MDL file must exist, since the existing header is reused.
  'sources' is a list of sources to write, if None, then model.sources is used.
  """
  if sources is None:
    sources = model.sources;
  dprintf(2,"writing %s model sources to NEWSTAR file\n",len(sources),filename);
  
  ra0,dec0 = model.fieldCenter();
  freq0 = freq0 or model.refFreq();
  # if freq0 is not specified, scan sources
  if freq0 is None:
    for src in sources:
      freq0 = (src.spectrum and getattr(src.spectrum,'freq0',None)) or getattr(src.flux,'freq0',None);
      if freq0:
        break;
    else:
      raise ValueError("unable to determine NEWSTAR model reference frequency, please specify one explicitly.");
  
  ff = open(filename,mode="wb");
  
  ### create GFH header
  gfh = numpy.zeros(512,dtype=numpy.uint8);
  datestr = time.strftime("%d-%m-%Y");
  timestr = time.strftime("%H:%M");
  struct.pack_into("4sii11s5s11s5si80sB",gfh,0,".MDL",512,1,
    datestr,timestr,datestr,timestr,0,
    os.path.splitext(os.path.basename(filename))[0],6);  # 6=datatype
  # link1/link2 gives the header size actually
  struct.pack_into("ii",gfh,152,512,512);
  gfh.tofile(ff);  
  
  # create MDH header
  mdh = numpy.zeros(64,dtype=numpy.uint8);
  struct.pack_into('iiii',mdh,12,1,576,0,2); # maxlin,pointer,num_sources,mtype
  struct.pack_into('f',mdh,28,getattr(model,'epoch',2000));
  struct.pack_into('ddd',mdh,32,ra0/(2*math.pi),dec0/(2*math.pi),freq0*1e-6);
  mdh.tofile(ff);

  # get the max ID, if specified
  max_id = max([ getattr(src,'newstar_id',0) for src in sources ]);
  # now loop over model sources
  # count how many are written out -- only point sources and gaussians are actually written out, the rest are skipped
  nsrc = 0;
  for src in sources:
    # create empty newstar source structure
    mdl = numpy.zeros(56,dtype=numpy.uint8);
    
    if src.shape and not isinstance(src.shape,ModelClasses.Gaussian):
      dprint(3,"skipping source '%s': non-supported type '%s'"%(src.name,src.shape.typecode));
      continue;
    
    stI = src.flux.I;
    # get l,m NCP position -- either from tag, or compute
    lm = getattr(src,'_lm_ncp',None);
    if lm:
      if isinstance(lm,(tuple,list)) and len(lm) == 2:
        l,m = lm;
      else:
        dprint(0,"warning: skipping source '%s' because its _lm_ncp attribute is malformed (tuple of 2 values expected)"%src.name);
        continue;
    else:
      l,m = radec_to_lm_ncp(ra0,dec0,src.pos.ra,src.pos.dec);
      
    # update source count
    nsrc += 1;
    # generate source id
    src_id = getattr(src,'newstar_id',None);
    if src_id is None:
      src_id = max_id = max_id+1;

    # encode position, flux, identifier -- also, convert flux from Jy to WU to Jy (1WU=5mJy)
    struct.pack_into('fffi',mdl,0,stI/0.005,l,m,src_id);
     
    # encode fractional polarization
    struct.pack_into('fff',mdl,16,*[ getattr(src.flux,stokes,0.0)/stI for stokes in "QUV" ]);
    
    ## encode flag & type bits
    ## Flag: bit 0= extended; bit 1= Q|U|V <>0 and no longer used according to Wim
    ## Type: bit 0= clean component; bit 3= beamed
    beamed = getattr(src,'flux_intrinsic',False) or getattr(src,'newstar_beamed',False);
    struct.pack_into('BB',mdl,52,
      1 if src.shape else 0,
      (1 if getattr(src,'newstar_cc',False) else 0) | (8 if beamed else 0));

    ### extended source parameters
    if src.shape:
      ## the procedure is NMOEXF in nscan/nmoext.for
      R0 = math.cos(src.shape.pa);
      R1 = -math.sin(src.shape.pa);
      R2 = (.5*src.shape.ex)**2;
      R3 = (.5*src.shape.ey)**2;
      ex = R2*R1*R1+R3*R0*R0          
      ey = R2*R0*R0+R3*R1*R1
      pa = 2*(R2-R3)*R0*R1
      struct.pack_into('fff',mdl,28,ex,ey,pa);

    ### spectral index
    if isinstance(src.spectrum,ModelClasses.SpectralIndex):
      struct.pack_into('f',mdl,40,src.spectrum.spi);
    
    if isinstance(src.flux,ModelClasses.PolarizationWithRM):
      struct.pack_into('f',mdl,44,src.flux.rm);
      
    mdl.tofile(ff);
    
  # update MDH header with the new number of sources
  struct.pack_into('i',mdh,20,nsrc);
  ff.seek(512);
  mdh.tofile(ff);
  ff.close();
  dprintf(1,"wrote %d sources to file %s\n",nsrc,filename);


Tigger.Models.Formats.registerFormat("NEWSTAR",load,"NEWSTAR model file",(".mdl",".MDL"),export_func=save);
